#! /usr/bin/env python3

"""Main XOInvader module, that is entry point to game.

Prepare environment for starting game and start it."""


import argparse
import logging

import xoinvader
from xoinvader import application
from xoinvader import constants


LOG = logging.getLogger(__name__)


def create_game():
    """Create XOInvader game instance."""

    import curses

    from xoinvader.curses_utils import Style

    # TODO: config-management: remove import order dependency
    # problem: importing of .common creates settings dict on import time
    #          then importing of game state pulls Ship and then Mixer
    #          mixer chooses mixer on import time based on settings
    #          update_system_settings called in run time, updates settings...
    #          but dependent entities already created when imported.
    # solutions:
    #          * Remove import time magic
    #            Create defaults, but make decisions on run time only
    #            Maybe this will lead to reorganizing code.

    app = application.get_application_class()()

    from xoinvader.ingame import InGameState
    from xoinvader.menu import MainMenuState, GameOverState

    Style().init_styles(curses)
    app.register_state(InGameState)
    app.register_state(MainMenuState)
    app.register_state(GameOverState)
    return app


def create_test_game():
    """Temporary function to create Pygame."""

    from xoinvader.teststate import TestState

    app = application.get_application_class()((800, 600), 0, 32)
    app.register_state(TestState)
    return app


def parse_args():
    """Parse incoming arguments."""
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-d", "--debug", action="store_true", help="enable debug mode")
    parser.add_argument(
        "-ns", "--no-sound", action="store_true", help="disable sounds")
    parser.add_argument(
        "-nc", "--no-color", action="store_true", help="disable colors")
    parser.add_argument(
        "-vd", "--video-driver", default=constants.DRIVER_NCURSES,
        type=str, choices=(constants.DRIVER_NCURSES, constants.DRIVER_SDL),
        help="use pygame")

    args = parser.parse_args()
    return args


def main():
    """Start the game!"""

    args = parse_args()
    # apply settings to engine settings
    xoinvader.init(args.__dict__)
    LOG.debug("Incoming args: %s", args)

    # TODO: backend-drivers; implement proper choosing
    if args.video_driver == constants.DRIVER_SDL:
        game = create_test_game()
    else:
        game = create_game()

    LOG.info("Staring game with '%s' driver", args.video_driver)
    return game.start()


if __name__ == "__main__":
    main()



"""Animation.

Animation is set of keyframes.
Value of selected attribute changes in time.

Keyframe:
  (time, value)

Objects have animation manager which manages animation graph and switching."""

from __future__ import division

from operator import itemgetter

from xoinvader.utils import Point, Timer


class AnimationBoundariesExceeded(Exception):
    """Exception to show that interpolated value will be incorrect."""

    def __init__(self, first, current_time, second):
        super(AnimationBoundariesExceeded, self).__init__(
            self,
            "Animation frame boundaries exceeded: {0} <= {1} <= {2}".format(
                first, current_time, second))


class InterpolationUnknownTypes(Exception):
    """Such type combination is unsupported."""

    def __init__(self, first, second):
        super(InterpolationUnknownTypes, self).__init__(
            self, "Unknown types of interpolating values: {0} and {1}".format(
                first, second))


# TODO: Implement animation graph and etc
class AnimationManager(object):
    """Manage list of object animation."""

    def __init__(self):
        self._animations = {}
        self._animation = None

    @property
    def animation(self):
        """AnimationManager's current animation name.

        To set animation - assign it's name.

        :getter: yes
        :setter: yes
        :type: str
        """

        if self._animation:
            return self._animation.name
        else:
            raise AttributeError("There is no available animation.")

    @animation.setter
    def animation(self, name):
        if name in self._animations:
            self._animation = self._animations[name]
        else:
            raise ValueError("No such animation: '{0}'.".format(name))

    def add(self, name, *args, **kwargs):
        """Add new animation, pass args to Animation class.

        See interface of `class::xoinvader.animation.Animation`.

        :param str name: animation name
        """

        animation = Animation(name, *args, **kwargs)
        self._animations[name] = animation

        if not self._animation:
            self._animation = animation

    def update(self):
        """Update manager's state."""

        if not self._animation:
            return

        try:
            self._animation.update()
        except StopIteration:
            return  # TODO: think about method to change animation


# pylint: disable=too-many-instance-attributes,too-many-arguments
# pylint: disable=too-few-public-methods
class Animation(object):
    """Animation unit.

    Animation object holds sorted list of (time, value) items and changes
    selected attribute of bound object according to local animation time.
    Time measured by timer. When current time is greater or equal then time
    of next keyframe - animation object changes it to appropriate value.
    When animation is done and if not looped - raise StopIteration.
    In case of interpolated animation value calculation occurs within two
    bounding frames and on frame switch.

    :param str name: animation name
    :param object bind: object to bind animation
    :param str attr: attribute to change in frames
    :param list keyframes: (float, object) tuples
    :param bool interp: interpolate values between frames or not
    :param bool loop: loop animation or not
    """

    def __init__(self, name, bind, attr, keyframes, interp=False, loop=False):
        self._name = name
        self._obj = bind
        self._attr = attr

        if not keyframes:
            raise ValueError("Animation keyframes must not be empty.")
        self._keyframes = sorted(keyframes, key=itemgetter(0))

        self._interp = interp
        self._loop = loop

        # Timer for tracking local time
        self._timer = Timer(self._keyframes[-1][0], lambda: True)
        self._timer.start()

        # Current keyframe index
        self._current = 0

        if self._interp:
            self.update = self._update_interpolated
        else:
            self.update = self._update_discrete

    @property
    def name(self):
        """Animation's name.

        :getter: yes
        :setter: no
        :type: str
        """
        return self._name

    def _apply_value(self, value):
        """Apply new value to linked object.

        :param obj value: value to apply
        """

        setattr(self._obj, self._attr, value)

    def _update_interpolated(self):
        """Advance animation and interpolate value.

        NOTE: animation frame switching depends on interp mode
        animation with interpolation switches frame only when current local
        time exceeds NEXT frames' time border.
        """

        self._check_animation_state()
        self._timer.update()

        current_time = self._timer.get_elapsed()
        keyframe = self._keyframes[self._current]
        next_keyframe = self._keyframes[self._current + 1]

        # it's time to switch keyframe
        if current_time >= next_keyframe[0]:
            self._current += 1
            keyframe = self._keyframes[self._current]

        if self._current == len(self._keyframes) - 1:
            self._apply_value(keyframe[1])
            self._current += 1
            self._check_animation_state()
            return

        next_keyframe = self._keyframes[self._current + 1]

        value = interpolate(keyframe, next_keyframe, current_time)
        self._apply_value(value)

    def _update_discrete(self):
        """Advance animation without interpolating value.

        NOTE: animation frame switching depends on interp mode
        discrete animation swiches frame and updates value only if
        current local time is >= time of current keyframe.
        No need to worry about calculating value between frames - thus
        no need to complicate behaviour.
        """

        self._check_animation_state()
        self._timer.update()

        keyframe = self._keyframes[self._current]

        # Check if animation need to switch keyframe
        if self._timer.get_elapsed() >= keyframe[0]:
            self._apply_value(keyframe[1])
            self._current += 1

    def _check_animation_state(self):
        """Check animation state and restart if needed.

        :raise StopIteration: when animation exceeded frames.
        """

        if len(self._keyframes) == self._current:
            if self._loop:
                self._current = 0
                self._timer.restart()
            else:
                self._timer.stop()
                raise StopIteration


def linear_equation(val1, val2, time1, time2, current_time):
    """Linear equation to get interpolated value.

    :param float val1: first keyframe value
    :param float val2: second keyframe value
    :param float time1: first keyframe local time
    :param float time2: second keyframe local time
    :param float current_time: current animation local time
    """

    return val1 + (val2 - val1) / (time2 - time1) * (current_time - time1)


def same_type(values, types):
    """Check if values are belongs to same type or type tuple.

    :param collections.Iterable values: values to check type similarity
    :param tuple|type types: type or tuple of types
    """

    return all(map(lambda it: isinstance(it, types), values))


def interpolate(first, second, current_time):
    """Interpolate value by two bounding keyframes.

    :param collections.Iterable first: first bounding keyframe
    :param collections.Iterable second: second bounding keyframe
    :param float current_time: current animation local time

    :raises AnimationBoundariesExceeded: when time interval is invalid
    :raises InterpolationUnknownTypes: when interpolating invalid types
    """

    if not first[0] <= current_time <= second[0]:
        raise AnimationBoundariesExceeded(first[0], current_time, second[0])

    def frames_of(*args):
        """If frames both of specified type."""
        return same_type((first[1], second[1]), args)

    if frames_of(int, float):
        value = linear_equation(
            float(first[1]), float(second[1]),
            float(first[0]), float(second[0]), float(current_time))

    elif frames_of(Point):
        value = linear_equation(
            first[1], second[1], float(first[0]), float(second[0]),
            float(current_time))
    else:
        raise InterpolationUnknownTypes(type(first[1]), type(second[1]))

    return value



"""Sound files handling."""

import logging

from six import add_metaclass

from xoinvader.common import Settings
from xoinvader.utils import Singleton


LOG = logging.getLogger(__name__)


def get_mixer():
    """Returns appropriate mixer class.

    :return: mixer implementation
    :rtype: class
    """

    choises = {
        True: DummyMixer,
        False: PygameMixer
    }

    return choises[Settings.system.no_sound]


# pylint: disable=no-self-use
class DummyMixer(object):
    """Mixer stub for no-sound mode."""

    def __init__(self):
        pass

    def register(self, *args, **kwargs):
        """Dummy register."""
        LOG.debug("Registering: %s; %s", args, kwargs)

    def play(self, *args, **kwargs):
        """Dummy play."""
        pass

    def mute(self):
        """Dummy mute."""
        LOG.debug("Muting.")

    def unmute(self):
        """Dummy unmute."""
        LOG.debug("Unmuting.")


@add_metaclass(Singleton)
class PygameMixer(object):
    """Handle sound files."""

    def __init__(self):
        # TODO: make 'sound' package alike 'application'
        # TODO: import appropriate sound class from separate files
        import pygame
        self.pygame = pygame  # Do not try it home

        self.pygame.mixer.init()

        self._sounds = {}
        self._mute = Settings.system.no_sound

    def register(self, object_id, sound_path):
        """Map object classname to sound object.

        :param str object_id: now this is class name
        :param str sound_path: path to object's sound
        """

        LOG.debug("Registering %s, %s.", object_id, sound_path)
        if object_id in self._sounds:
            return

        sound = self.pygame.mixer.Sound(sound_path)
        self._sounds.update({object_id: sound})

    def play(self, object_id, *args, **kwargs):
        """Play sound object."""

        if self._mute:
            return

        if object_id in self._sounds:
            self._sounds[object_id].play(*args, **kwargs)

    def mute(self):
        """Mute."""
        self._mute = True

    def unmute(self):
        """Unmute."""
        self._mute = False


Mixer = get_mixer()  # pylint: disable=invalid-name



"""Base class for game application.

Only one application can exist in one interpreter instance.
"""

from tornado import ioloop

from xoinvader.constants import DEFAULT_FPS, DRIVER_NCURSES, DRIVER_SDL
from xoinvader.common import Settings


class ApplicationNotInitializedError(Exception):
    """Raise when try to get not initialized application."""

    def __init__(self):
        super(ApplicationNotInitializedError, self).__init__(
            "Application not initialized.")


def get_current():
    """Current application getter.

    :return: current application object
    """

    return Application.current()


# TODO: implement proper choosing by env
def get_application_class():
    """Application class getter.

    :return: application class based on environment
    """

    driver_map = {
        DRIVER_NCURSES: get_ncurses_application,
        DRIVER_SDL: get_pygame_application,
    }

    return driver_map[Settings.system.video_driver]()


def get_ncurses_application():
    """Incapsulate ncurses-related stuff.

    :return: CursesApplication class
    """

    from .ncurses_app import CursesApplication

    return CursesApplication


def get_pygame_application():
    """Incapsulate pygame-related stuff.

    :return: PygameApplication class
    """

    from .pygame_app import PygameApplication

    return PygameApplication


class Application(object):
    """Base application class for backend-specific application classes.

    Provides state primitive mechanism and some useful getters/setters.
    """

    __CURRENT_APPLICATION = None
    """Instance of the current application."""

    def __init__(self):
        Application.__CURRENT_APPLICATION = self

        self._state = None
        self._states = {}
        self._screen = None
        self._fps = DEFAULT_FPS
        self._ioloop = ioloop.IOLoop.instance()

        self._pc = ioloop.PeriodicCallback(self.tick, 30, self._ioloop)
        self._pc.start()

    def __del__(self):
        Application.__CURRENT_APPLICATION = None

    def tick(self):
        """Callback to execute at every frame."""

        self._state.events()
        self._state.update()
        self._state.render()

    @classmethod
    def current(cls):
        """Return the current application if exists, raise otherwise."""

        if cls.__CURRENT_APPLICATION:
            return cls.__CURRENT_APPLICATION
        else:
            raise ApplicationNotInitializedError()

    @property
    def state(self):
        """Current state.

        :getter: Return current state
        :setter: Set current state
        :type: :class:`xoinvader.state.State`
        """
        if self._state:
            return self._state
        else:
            raise AttributeError("There is no available state.")

    @state.setter
    def state(self, name):
        """Setter."""
        if name in self._states:
            self._state = self._states[name]
        else:
            raise KeyError("No such state: '{0}'.".format(name))

    @property
    def states(self):
        """State names to State classes mapping.

        :getter: yes
        :setter: no
        :type: dict
        """
        return self._states

    def register_state(self, state):
        """Add new state and initiate it with owner.

        :param state: state class to register
        :type state: :class:`xoinvader.state.State`
        """
        name = state.__name__
        state_object = state(self)
        self._states[name] = state_object
        if len(self._states) == 1:
            self._state = self._states[name]

        # NOTE: State cannot instantiate in State.__init__ objects that
        #       want access to state because there is no instance at creation
        #       moment. For such objects state can declare it's 'postinit'
        #       method.
        state_object.postinit()

    def deregister_state(self, name):
        """Remove existing state.

        :param str name: name of state
        """

        state = self._states.pop(name)
        del state

    def trigger_state(self, state, *args, **kwargs):
        """Change current state and pass args and kwargs to it.

        :param str state: state name
        """

        self.state = state
        self.state.trigger(*args, **kwargs)

    @property
    def fps(self):
        """Frames per second.

        :getter: yes
        :setter: yes
        :type: int
        """
        return self._fps

    @fps.setter
    def fps(self, val):
        """Setter."""
        self._fps = int(val)

    @property
    def screen(self):
        """Application's screen Surface.

        :getter: yes
        :setter: no
        :type: class::`curses.Window` or class::`pygame.display.Surface`
        """
        return self._screen

    def start(self):
        """Start main application loop."""

        if not self._state:
            raise AttributeError("There is no available state.")

        self._ioloop.start()

    def stop(self):
        """Stop application."""

        self._pc.stop()
        self._ioloop.add_callback(self._ioloop.stop)



"""Scoreboard routines.

Scoreboard file is CSV of username,score rows.
If file doesn't exist it will be recreated from default stub.
"""

import operator
import os
import csv

from xoinvader.common import Settings


def items():
    """Return all scorefile entries.

    :return [(str, int)]: (username, score) pairs
    """

    return _load()


def add(username, score):
    """Add new entry to scoreboard file, sorted by score.

    :param str username: username
    :param int score: user score
    """

    scores = items()
    scores.append((username, int(score)))
    scores.sort(key=operator.itemgetter(1))

    _save(scores)


def lowest():
    """Return lowest result in game.

    :return int: lowest score
    """

    scores = items()
    if scores:
        return min(scores, key=operator.itemgetter(1))[1]
    else:
        return 0


def highest():
    """Return highest result in game.

    :return int: highest score
    """

    scores = items()
    if scores:
        return max(scores, key=operator.itemgetter(1))[1]
    else:
        return 0


def _load():
    """Load scores from scorefile.

    :return [(str, int)]: scores
    """

    scores = []

    try:
        with open(Settings.path.scoreboard) as scorefile:
            for entry in csv.reader(scorefile):
                try:
                    name, score = entry
                    scores.append((name, int(score)))
                except ValueError:
                    # probably CSV is corrupted, skip failure entries
                    pass
    except IOError:
        pass

    return scores


def _save(scores):
    """Save scores to scorefile.

    :param [(str, int)] scores: scores to save
    """

    dirname = os.path.dirname(Settings.path.scoreboard)

    if not os.path.isdir(dirname):
        os.mkdir(dirname)

    with open(Settings.path.scoreboard, "w") as scorefile:
        csv.writer(scorefile).writerows(scores)



import uuid


#pylint: disable=all
class Entity(object):
    Catalog = {}
    __slots__ = ("name", "uid", "components")

    def __new__(cls, name, uid=None):
        if name not in cls.Catalog:
            entity = super(Entity, cls).__new(cls)
            cls.Catalog[name] = entity
        else:
            entity = cls.Catalog[name]
        return entity

    def __hash__(self):
        return hash(self.uid)

    def __init__(self, name="", uid=None):
        self.name = name
        self.uid = uid or uuid.uuid4()
        self.components = {}

    def __repr__(self):
        cname = self.__class__.__name__
        name = self.name or self.uid
        if name != self.uid:
            name = "{0}:{1}".format(name, self.uid)
        return "<{0} {1}>".format(cname, name)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.uid == other.uid
        elif isinstance(other, self.uid.__class):
            return self.uid == other
        return False

    def __str__(self):
        return str(self.components)

    def __getitem__(self, key):
        return self.components[key]

    def __setitem__(self, key, val):
        if isinstance(val, Component):
            self.components[key] = val


class Component(object):
    pass


class System(object):
    pass



# -*- coding: utf-8 -*-
""" Graphical user interface widgets."""


from __future__ import unicode_literals

# TODO: make working with styles pretty
from xoinvader.curses_utils import Style
from xoinvader.render import Renderable
from xoinvader.utils import Surface, Timer


class TextWidget(Renderable):
    """Simple text widget.

    :param pos: widget's global position
    :type pos: `xoinvader.utils.Point`

    :param text: contained text
    :type text: string

    .. note:: add [ [style], ...] support

    :param style: curses style for text
    :type style: integer(curses style)
    """

    render_priority = 1
    draw_on_border = True

    def __init__(self, pos, text, style=None):
        self._pos = pos
        self._text = text
        self._style = style
        self._image = self._make_image()

    def _make_image(self):
        """Make Surface object from text and style.

        :return: Surface instance
        :rtype: `xoinvader.utils.Surface`
        """

        _style = self._style or Style().gui["normal"]
        return Surface(
            [[ch for ch in self._text]],
            [[_style for _ in range(len(self._text))]])

    def update(self, text=None, style=None):
        """Obtain (or not) new data and refresh image.

        :param text: new text
        :type: string

        :param style: new style
        :type: integer(curses style)
        """

        if text:
            self._text = text
        if style:
            self._style = style
        if text or style:
            self._image = self._make_image()

    def get_render_data(self):
        return [self._pos], self._image.get_image()


# TODO XXX FIXME: [proper-gui-hierarchy]
#                 This bloody mess bourned in hell suburbans to serve
#                 to one great target: easily enable updating by callback
#                 to update score string. It was so easy to do...
#                 Reimplement TextWidget to support callbacks too.
class TextCallbackWidget(TextWidget):
    """Simple text widget with callback.

    :param class::`xoinvader.utils.Point` pos: widget's global position

    :param str text: contained text

    .. Note:: add [ [style], ...] support

    :param int style: curses style for text
    """

    def __init__(self, pos, callback, style=None):

        self._callback = callback

        super(TextCallbackWidget, self).__init__(
            pos, callback(), style)

    def update(self):
        self._text = self._callback()
        self._image = self._make_image()


class MenuItemWidget(TextWidget):
    """Selectable menu item widget.

    :param pos: widget's global position
    :type pos: `xoinvader.utils.Point`

    :param text: contained text
    :type text: string

    :param template: left and right markers
    :type template: tuple of two strings

    .. note:: add [ [style] ... ] support

    :param style: curses style for text
    :type style: integer(curses style)
    """

    render_priority = 1

    def __init__(self, pos, text, template=("* ", " *"), style=None):
        self._left = template[0]
        self._right = template[1]
        self._selected = False

        super(MenuItemWidget, self).__init__(pos, text, style)

    def _make_image(self):
        """Make Surface object from text, markers and style.

        :return: Surface instance
        :rtype: `xoinvader.utils.Surface`
        """

        _style = self._style or Style().gui["yellow"]
        if self._selected:
            _full_text = "".join([self._left, self._text, self._right])
        else:
            _full_text = self._text

        return Surface(
            [[ch for ch in _full_text]],
            [[_style for _ in range(len(_full_text))]])

    def toggle_select(self):
        """Draw or not selector characters."""
        self._selected = not self._selected

    def select(self):
        """Select and refresh image."""

        self._selected = True
        self._image = self._make_image()

    def deselect(self):
        """Deselect and refresh image."""

        self._selected = False
        self._image = self._make_image()

    @property
    def selected(self):
        """Shows is item selected or not.

        .. warning:: Complete menu workflow.

        :getter: yes
        :setter: no
        :type: boolean
        """
        return self._selected

    def get_render_data(self):
        return [self._pos], self._image.get_image()


# pylint: disable=too-many-arguments
class PopUpNotificationWidget(TextWidget):
    """Widget that allows to show short messages with timeout.

    .. warning:: Experimental stuff. Fix constructor 6/5.

    :param pos: global position
    :type pos: :class:`xoinvader.utils.Point`

    :param text: text for display
    :type text: string

    :param style: curses style
    :type style: int | [int]

    :param timeout: timer timeout
    :type timeout: float

    :param callback: callback for removal object
    :type callback: function
    """

    def __init__(self, pos, text, style=None, timeout=1.0, callback=None):
        super(PopUpNotificationWidget, self).__init__(pos, text, style)

        self._callback = callback
        self._timer = Timer(timeout, self._finalize_cb)
        self._update_text = super(PopUpNotificationWidget, self).update
        self._timer.start()

    def _finalize_cb(self):
        """Finalize callback, e.g. pass to it self for removal."""

        if self._callback:
            self._callback(self)

    def update(self, text=None, style=None):
        self._update_text(text, style)
        self._timer.update()


class WeaponWidget(Renderable):
    """Widget for displaying weapon information.

    .. warning:: !!! Duplicates TextWidget !!!

    :param pos: global position
    :type pos: :class:`xoinvader.utils.Point`

    :param get_data: callback for getting data
    :type get_data: function
    """

    render_priority = 1
    draw_on_border = True

    def __init__(self, pos, get_data):
        self._pos = pos
        self._get_data = get_data
        self._data = self._get_data()
        self._image = self._make_image()

    def _make_image(self):
        """Return Surface object."""

        return Surface(
            [[ch for ch in self._data]],
            [[Style().gui["yellow"] for _ in range(len(self._data))]])

    def update(self):
        """Obtain new data and refresh image."""

        self._data = self._get_data()
        self._image = self._make_image()

    def get_render_data(self):
        """Return render specific data."""
        return [self._pos], self._image.get_image()


# pylint: disable=too-many-instance-attributes
class Bar(Renderable):
    """Progress bar widget.

    :param pos: Bar's global position
    :type pos: `xoinvader.utils.Point`

    :param str prefix: text before the bar
    :param str postfix: text after the bar
    :param str left: left edge of the bar
    :param str right: right edge of the bar
    :param str marker: symbol that fills the bar
    :param int marker_style: curses style for marker (passes to render)
    :param str empty: symbols that fills empty bar space (without marker)
    :param int  empty_style: curses style for empty marker (passes to render)
    :param int count: number of markers in the bar
    :param int maxval: max value of displayed parameter (affects the accuracy)
    :param int general_style: style of other characters(prefix, postfix, etc)

    :param stylemap: mapping of compare functions and integers to curses style
    :type stylemap: dict(function, integer(curses style)

    :param function callback: calls if not None to get new percentage value
    """

    render_priority = 1
    draw_on_border = True

    def __init__(
            self, pos,
            prefix="", postfix="",
            left="[", right="]",
            marker="█", marker_style=None,
            empty="-", empty_style=None,
            count=10, maxval=100,
            general_style=None,
            stylemap=None, callback=None
    ):

        self._pos = pos
        self._prefix = prefix
        self._postfix = postfix
        self._left = left
        self._right = right
        self._marker = marker
        self._marker_style = marker_style
        self._empty = empty
        self._empty_style = empty_style
        self._count = count
        self._maxval = maxval
        self._general_style = general_style
        self._stylemap = stylemap
        self._callback = callback

        self._template = "".join([
            str(val) for val in [
                self._prefix,
                self._left, "{blocks}", self._right,
                self._postfix
            ]
        ])

        self._current_count = self._count
        self._image = None
        self._update_image()

    def _update_current_count(self, val):
        """Normalize current percentage and update count of marker blocks.

        :param int val: value to normalize
        """
        self._current_count = int(round(val * self._count / self._maxval))

    def _style(self, val):
        """Return style in depend on percentage."""

        for cmp_func, bar_style in self._stylemap.items():
            if cmp_func(val):
                return bar_style
        return None

    def _update_image(self):
        """Update image in depend on percentage."""

        left = self._marker * self._current_count
        right = self._empty * (self._count - self._current_count)
        bar = self._template.format(blocks=left + right)
        image = []
        for char in bar:
            if char == self._marker:
                image.append((char, self._marker_style))
            else:
                image.append((char, self._general_style))
        self._image = Surface(
            [[ch[0] for ch in image]],
            [[st[1] for st in image]])

    def update(self, val=None):
        """Update bar if there's need for it."""

        if self._callback:
            val = self._callback()

        if val is None:
            raise ValueError("val = None, what to do?")

        self._marker_style = self._style(val)
        self._update_current_count(val)
        self._update_image()

    def get_render_data(self):
        return [self._pos], self._image.get_image()



"""Pygame helper module."""

import pygame


# pylint: disable=no-member
pygame.init()


# TODO: maybe there is need to rewrite this, or completely remove.
def create_display(res, fullscreen, depth, caption):
    """Create pygame display."""

    display = pygame.display.set_mode(res, fullscreen, depth)
    pygame.display.set_caption(caption)
    return display


def get_clock():
    """Helper for unification with other backends."""
    return pygame.time.Clock()



"""Test xoinvader.level module."""

from xoinvader.level import Level


# pylint: disable=invalid-name,protected-access,missing-docstring
def test_level():
    e = Level()

    # pylint: disable=too-few-public-methods
    class MockObject(object):
        def __init__(self):
            self.value = 0

        def add(self):
            self.value += 10

    a = MockObject()
    b = MockObject()

    e.add_event(10, a.add)
    e.add_event(20, b.add)

    e.speed = 10
    assert e.speed == 10
    e.update()
    assert a.value == 0
    assert b.value == 0
    e.start()
    e.update()
    assert a.value == 10
    assert b.value == 0
    e.update()
    assert a.value == 10
    assert b.value == 10

    assert not e.running



"""Test xoinvader.curses_utils module."""

import time

from xoinvader.curses_utils import Clock, get_clock
from xoinvader.utils import isclose


# pylint: disable=invalid-name
def test_clock():
    """xoinvader.curses_utils.Clock"""

    assert isinstance(get_clock(), Clock)

    clock = Clock()
    t1 = int(time.perf_counter())
    assert clock.get_time() == 0
    # assert clock.get_fps() == 0.0
    ticked = clock.tick()
    t2 = int(time.perf_counter())
    assert isclose(t2 - t1, ticked)
    assert clock.get_time() == ticked



"""Test related utilites."""

from xoinvader.state import State


PREFIX = "xoinvader/tests/fixtures/"
"""Test fixture data prefix."""


# pylint: disable=missing-docstring
class StateMock(State):
    """Mock State interface."""

    def __init__(self, owner):
        super(StateMock, self).__init__(owner)
        self.loop_count = 1

        def nop():
            pass

        # Hooks for loop testing
        self.on_events = nop
        self.on_update = nop
        self.on_render = nop

    def events(self):
        self.on_events()

    def update(self):
        self.on_update()

    def render(self):
        self.on_render()
        if self.loop_count <= 0:
            self.owner.stop()

        self.loop_count -= 1


class AnotherStateMock(StateMock):
    """Class for test with two instances."""
    pass



"""Tests for xoinvader.scoreboard module."""

from operator import itemgetter
import os

import pytest

from xoinvader import scoreboard
from xoinvader.tests.common import PREFIX


SCOREBOARD_DATA = [
    ("most", 1700),
    ("most", 1440),
    ("most", 0),
]
"""Scoreboard test data."""

SCOREBOARD_VALID = os.path.join(PREFIX, "scoreboard_valid")
"""Contains scoreboard valid data."""

SCOREBOARD_EMPTY = os.path.join(PREFIX, "scoreboard_empty")
"""Empty scoreboard file."""

SCOREBOARD_NONEXISTENT = os.path.join(PREFIX, "scoreboard_nonexistent")
"""No such file."""

SCOREBOARD_CORRUPTED = os.path.join(PREFIX, "scoreboard_corrupted")
"""Corrupted, but recoverable scoreboard file."""


# pylint: disable=redefined-outer-name
@pytest.fixture()
def mock_scorepath(monkeypatch):
    """Fixture to mock Settings.path.scoreboard entry and revert it back."""

    def inner(path):
        """Inner routine.

        :param str path: new path to set
        """

        mocked = scoreboard.Settings.fullcopy()
        mocked.path.scoreboard = path
        monkeypatch.setattr(scoreboard, "Settings", mocked)

    return inner


@pytest.mark.parametrize(("path", "expected"), (
    (SCOREBOARD_EMPTY, []),
    (SCOREBOARD_NONEXISTENT, []),
    (SCOREBOARD_CORRUPTED, SCOREBOARD_DATA),
    (SCOREBOARD_VALID, SCOREBOARD_DATA),
))
def test_items(path, expected, mock_scorepath):
    """xoinvader.scoreboard.items()."""

    mock_scorepath(path)
    assert scoreboard.items() == expected
    assert scoreboard._load() == expected


def test_add(tmpdir, mock_scorepath):
    """xoinvader.scoreboard.add()."""

    mock_scorepath(str(tmpdir.mkdir("data").join("scoreboard")))
    scoreboard.add("test", 1000)
    assert ("test", 1000) in scoreboard.items()

    scoreboard.add("test", 2000)
    items = scoreboard.items()
    assert ("test", 1000) in items
    assert ("test", 2000) in items


def test_lowest(mock_scorepath):
    """xoinvader.scoreboard.lowest()."""

    mock_scorepath(SCOREBOARD_VALID)
    assert scoreboard.lowest() == min(SCOREBOARD_DATA,
                                      key=itemgetter(1))[1]

    mock_scorepath(SCOREBOARD_EMPTY)
    assert scoreboard.lowest() == 0


def test_highest(mock_scorepath):
    """xoinvader.scoreboard.highest()."""

    mock_scorepath(SCOREBOARD_VALID)
    assert scoreboard.highest() == max(SCOREBOARD_DATA,
                                       key=itemgetter(1))[1]

    mock_scorepath(SCOREBOARD_EMPTY)
    assert scoreboard.highest() == 0


def test__save(tmpdir, mock_scorepath):
    """xoinvader.scoreboard._save()."""

    datadir = tmpdir.mkdir("data")
    scorefile = datadir.join("scoreboard")
    # Check that file doesn't exist
    assert not scorefile.check()

    mock_scorepath(str(scorefile))

    scoreboard._save(SCOREBOARD_DATA)
    assert scorefile.check()
    assert scoreboard.items() == SCOREBOARD_DATA

    # scoreboard._save() also must create directory for the scoreboard file
    datadir.remove()
    assert not scorefile.check()
    scoreboard._save(SCOREBOARD_DATA)
    assert scorefile.check()



"""xoinvader.application.Application unit test."""

import pytest

from xoinvader import constants
from xoinvader.common import Settings
import xoinvader.curses_utils
from xoinvader.application import (
    get_application_class, get_current, Application,
    ApplicationNotInitializedError
)
from xoinvader.application.ncurses_app import CursesApplication

from xoinvader.tests.common import StateMock, AnotherStateMock


def test_application(monkeypatch):
    """xoinvader.application.Application"""

    with pytest.raises(ApplicationNotInitializedError):
        get_current()

    app = Application()
    assert isinstance(get_current(), Application)
    assert pytest.raises(AttributeError, lambda: app.state)
    assert pytest.raises(AttributeError, app.start)
    assert app.screen is None
    app.fps = 40
    assert app.fps == 40

    # One element
    app.register_state(StateMock)
    assert len(app._states) == 1  # pylint: disable=protected-access
    assert isinstance(app.state, StateMock)

    assert pytest.raises(KeyError, lambda: setattr(app, "state", "test"))

    # Many elements
    app.register_state(AnotherStateMock)
    assert len(app.states) == 2
    # Ensure that Application._state hasn't changed
    assert isinstance(app.state, StateMock)

    app.state = AnotherStateMock.__name__
    assert isinstance(app.state, AnotherStateMock)

    # Test Application.trigger_state
    called_with = []

    def new_trigger(self, *args, **kwargs):
        called_with.extend([args, kwargs])

    monkeypatch.setattr(StateMock, "trigger", new_trigger)
    app.trigger_state(StateMock.__name__, 10, test=20)
    assert called_with == [(10,), {"test": 20}]


def test_curses_application(monkeypatch):
    """xoinvader.application.CursesApplication"""

    monkeypatch.setattr(
        xoinvader.curses_utils, "create_window",
        lambda *args, **kwargs: True)
    monkeypatch.setattr(
        xoinvader.curses_utils, "deinit_curses",
        lambda *args, **kwargs: True)

    assert CursesApplication == get_application_class()
    app = CursesApplication()
    assert isinstance(get_current(), CursesApplication)

    assert app.set_caption("test") is None
    assert _test_application_loop(app)


def test_pygame_application(monkeypatch):
    """xoinvader.application.PygameApplication"""

    pygame = pytest.importorskip("pygame")

    from xoinvader.application.pygame_app import PygameApplication

    monkeypatch.setattr(pygame, "init", lambda: (0, 6))
    monkeypatch.setattr(pygame, "quit", lambda: True)
    monkeypatch.setattr(pygame.display, "set_mode", lambda *a: True)
    monkeypatch.setattr(pygame.display, "update", lambda: True)
    monkeypatch.setattr(pygame.key, "set_repeat", lambda *a: True)

    Settings.system.video_driver = constants.DRIVER_SDL

    # Empty object
    assert PygameApplication == get_application_class()
    app = PygameApplication((800, 600), 0, 32)
    app.set_caption("test")
    assert isinstance(get_current(), PygameApplication)
    assert app.screen is not None
    assert pytest.raises(ValueError, lambda: setattr(app, "fps", "thirty"))
    assert pytest.raises(AttributeError, lambda: app.state)

    assert _test_application_loop(app)


def _test_application_loop(app):
    """Test main loop behaviour."""

    assert pytest.raises(AttributeError, app.start)
    app.register_state(StateMock)
    assert app.start() is None

    app.register_state(AnotherStateMock)
    app.state = AnotherStateMock.__name__
    assert app.start() is None
    assert app.stop() is None
    return True



"""Pygame application class."""

import pygame

from xoinvader.application import Application


# pylint: disable=no-member
class PygameApplication(Application):
    """Pygame powered application backend.

    :param resolution: surface resolution
    :type resolution: (int, int)

    :param flags: collection of additional options
    :type flags: int (bitmask)

    :param depth: number of bits to use for color
    :type depth: int
    """

    def __init__(self, resolution=(0, 0), flags=0, depth=0):
        super(PygameApplication, self).__init__()

        self._screen = pygame.display.set_mode(resolution, flags, depth)
        pygame.key.set_repeat(50, 50)

    def set_caption(self, caption, icontitle=""):
        """Set window caption.

        :param caption: window caption
        :type caption: str

        :param icontitle: short title
        :type icontitle: str
        """

        if self._screen:
            pygame.display.set_caption(caption, icontitle)

    def stop(self):
        self._ioloop.add_callback(pygame.quit)
        super(PygameApplication, self).stop()



"""Base handler class."""


# pylint: disable=too-few-public-methods
class Command(object):
    """Base command class."""

    def execute(self, actor):
        """Execute command with actor.

        :param actor: actor
        :type: actor: object

        .. note:: Make actor base class.
        """
        raise NotImplementedError


class Handler(object):
    """
    Base game handler.

    Provides accessing to main State's fields and handle function stub.

    :param owner: handler's owner
    :type owner: :class:`xoinvader.state.State`
    """

    def __init__(self, owner):
        self._owner = owner
        self._screen = owner.screen
        self._actor = owner.actor

    def handle(self):
        """Handle event."""
        raise NotImplementedError



import pytest

from xoinvader.vector import Vec2, Vec3


def test_point_operations():
    ax, ay, az = 10, 10, 10
    bx, by, bz = 20, 20, 20
    a = Vec3(ax, ay, az)
    b = Vec3(bx, by, bz)

    assert a.x == ax
    assert a.y == ay
    assert a.z == az
    assert b.x == bx
    assert b.y == by
    assert b.z == bz

    assert repr(a) == "Vec3(x={0}, y={1}, z={2})".format(
        a.x, a.y, a.z)

    assert a + b == Vec3(ax + bx, ay + by, az + bz)
    assert a - b == Vec3(ax - bx, ay - by, az - bz)
    assert a + 5 == Vec3(ax + 5, ay + 5, az + 5)
    assert a - 5 == Vec3(ax - 5, ay - 5, az - 5)
    assert a * 5 == Vec3(ax * 5, ay * 5, az * 5)
    assert a / 5 == Vec3(ax / 5, ay / 5, az / 5)

    with pytest.raises(ValueError):
        assert a + "a"
    with pytest.raises(ValueError):
        assert a - "a"
    with pytest.raises(ValueError):
        assert a * "a"
    with pytest.raises(ValueError):
        assert a / "a"
    with pytest.raises(ValueError):
        assert a == "a"

    a.x = bx
    a.y = by
    a.z = bz

    assert a.x == bx
    assert a.y == by
    assert a.z == bz

    b.x = -bx
    b.y = -by
    b.z = -bz

    assert a + b == Vec3(0, 0, 0)
    assert a + Vec3(-50, -50, -50) == Vec3(-30, -30, -30)

    assert Vec3(1.9, 1.9)[int] == Vec3(1, 1, 0)



"""Test xoinvader.background module."""

from copy import copy
import os

import pytest

from xoinvader.background import Background, Chunk, load_chunks
from xoinvader.common import Settings
from xoinvader.tests.common import PREFIX
from xoinvader.utils import Point


CHUNK_NO_NAME = os.path.join(PREFIX, "chunk_no_name.bg")
"""Background file contains unnamed chunk."""

CHUNK_DUPLICATE_NAMES = os.path.join(PREFIX, "chunk_duplicate_names.bg")
"""Background file contains duplicated chunk names."""

CHUNK_NO_CHUNKS = os.path.join(PREFIX, "no_chunks.bg")
"""Background file doesn't contain chunks at all."""

CHUNK_NORMAL = os.path.join(PREFIX, "chunk_normal.bg")
"""Absolutely valid background file."""


# pylint: disable=invalid-name,protected-access,missing-docstring
def test_chunk():
    c = Chunk("test")
    assert not c.lines
    assert len(c) == 0  # pylint: disable=len-as-condition
    assert c.name == "test"

    c.add_line("wut")
    assert c.lines[0] == "wut"
    assert len(c) == 1

    c[0] = "ugh"
    assert c[0] == "ugh"


def test_load_chunks():
    with pytest.raises(ValueError):
        load_chunks(CHUNK_NO_NAME)

    with pytest.raises(ValueError):
        load_chunks(CHUNK_DUPLICATE_NAMES)

    with pytest.raises(ValueError):
        load_chunks(CHUNK_NO_CHUNKS)

    c = load_chunks(CHUNK_NORMAL)
    assert len(c) == 3
    assert c[0][0] == "qweqwe"
    assert len(c[0]) == 1

    d = load_chunks(CHUNK_NORMAL, 3)
    assert d[0][0] == "qwe"


# pylint: disable=too-many-statements
def test_background():
    Settings.layout.field.edge.x = 3
    Settings.layout.field.edge.y = 2

    b = Background()
    assert not b.loop
    assert not b.loop_all
    assert not b.speed
    assert not b.chunks
    assert not b.background
    assert not b._background_surface
    assert not b._current_chunk
    assert not b._current_chunk_num
    assert not b._chunk_line
    assert not b._ticks_since_last_update

    b.background = ["qwe", "asd"]
    assert b.background == ["qwe", "asd"]

    b.chunks = "test"
    assert b.chunks == "test"

    assert b._advance_chunk(1) == "   "

    b = Background(CHUNK_NORMAL)
    assert len(b.chunks) == 3
    assert len(b.chunks[2]) == 3

    b.start()
    assert b._current_chunk is b.chunks[0]
    assert b.background == [
        "   ",
        "   "
    ]

    b.start(True)
    assert b._chunk_line == 1
    assert b._current_chunk is b.chunks[1]
    assert b.background == [
        "qwe",
        "asd"
    ]

    assert b._chunk_line == 1
    assert b._advance_chunk(1) == "zxc"
    assert b._chunk_line == 1
    assert b._advance_chunk(-1) == "123"
    assert b._chunk_line == 0
    assert b._advance_chunk(1) == "zxc"
    assert b._chunk_line == 1
    b.loop = True
    assert b._advance_chunk(1) == "123"
    assert b._advance_chunk(1) == "!@#"
    assert b._advance_chunk(-1) == "zxc"
    assert b._advance_chunk(-1) == "!@#"
    b.loop = False
    b.loop_all = True
    assert b._advance_chunk(1) == "123"
    assert b._advance_chunk(1) == "!@#"
    assert b._advance_chunk(1) == "qwe"
    assert b._advance_chunk(1) == "asd"
    assert b._advance_chunk(1) == "zxc"
    assert b._advance_chunk(-1) == "123"
    assert b._advance_chunk(-1) == "zxc"
    assert b._advance_chunk(-1) == "asd"
    assert b._advance_chunk(-1) == "qwe"
    assert b._advance_chunk(-1) == "!@#"
    b.loop_all = False
    assert b._advance_chunk(1) == "123"
    assert b._advance_chunk(1) == "!@#"
    assert b._advance_chunk(1) == "   "

    pos, gen = b.get_render_data()
    assert pos == [Point(0, 0)]
    assert next(gen) == (Point(0, 0, 0), "q", None)
    assert next(gen) == (Point(1, 0, 0), "w", None)
    assert next(gen) == (Point(2, 0, 0), "e", None)
    assert next(gen) == (Point(0, 1, 0), "a", None)
    assert next(gen) == (Point(1, 1, 0), "s", None)
    assert next(gen) == (Point(2, 1, 0), "d", None)
    with pytest.raises(StopIteration):
        next(gen)

    assert not b.update()
    b.speed = 40
    bg = copy(b.background)
    b.update()
    assert bg == b.background
    b.update()
    assert not bg == b.background

    b.speed = -40
    bg = copy(b.background)
    b.update()
    assert bg == b.background
    b.update()
    assert not bg == b.background



"""MainMenuState-related input and event handlers."""

from xoinvader.gui import (
    TextCallbackWidget, TextWidget, MenuItemWidget, PopUpNotificationWidget
)
from xoinvader.keys import K_ENTER, K_ESCAPE, K_R, K_N
from xoinvader.state import State
from xoinvader.utils import Point
from xoinvader.render import render_objects
from xoinvader.handlers import Handler


# pylint: disable=missing-docstring
def exit_game_command(actor):
    actor.owner.stop()


def to_ingame_command(actor):
    actor.owner.state = "InGameState"


def show_popup(actor):
    actor.notify("some_text")


# pylint: disable=too-few-public-methods
class MainMenuInputHandler(Handler):

    def __init__(self, owner):
        super(MainMenuInputHandler, self).__init__(owner)

        self._command_map = {
            K_ESCAPE: to_ingame_command,
            K_R: exit_game_command,
            K_N: show_popup
        }

    def handle(self):
        key = self._screen.getch()
        command = self._command_map.get(key)
        if command:
            if command in [exit_game_command, to_ingame_command, show_popup]:
                command(self._owner)
            else:
                command(self._actor)


# pylint: disable=too-few-public-methods
class MainMenuEventHandler(Handler):

    def __init__(self, owner):
        super(MainMenuEventHandler, self).__init__(owner)

        self._input_handler = MainMenuInputHandler(owner)

    def handle(self):
        self._input_handler.handle()
        # Some other event stuff


# Compound renderable object
# Contains renderables as InGame state contains.
# Or PlayerShip object contains weapons.
class Menu(object):
    """
    Represents menu.

    Keeps MenuItemWidget : owner.method mapping.
    Actor of any game menu state.
    """

    def __init__(self, items=None):
        self._items = items if items else list()

    def move_up(self):
        pass

    def move_down(self):
        pass


# recursion?
class SubMenu(Menu):
    pass


class MainMenuState(State):

    def __init__(self, owner):
        super(MainMenuState, self).__init__(owner)
        self._screen = owner.screen
        self._actor = None  # Should be some of Menu instances?

        self._objects = [
            TextWidget(Point(4, 4), "Whoooch"),
            MenuItemWidget(Point(10, 10), "First menu item"),
            MenuItemWidget(Point(10, 11), "Second menu item")
        ]
        self._objects[1].select()

        self._current_menu = None
        self._events = MainMenuEventHandler(self)

    def notify(self, text, pos=Point(15, 15)):
        self._objects.append(
            PopUpNotificationWidget(
                pos, text or "ololo",
                callback=self._objects.remove))

#   def register_menu_item(self, caption, item_action_list):
    def events(self):
        self._events.handle()

    def update(self):
        for o in self._objects:
            o.update()

    def render(self):
        self._screen.erase()
        self._screen.border(0)
        # render_objects(self._actor.get_objects, self._screen)
        render_objects(self._objects, self._screen)


# pylint: disable=too-few-public-methods
class GameOverInputHandler(Handler):

    def __init__(self, owner):
        super(GameOverInputHandler, self).__init__(owner)

        self._command_map = {
            K_ESCAPE: exit_game_command,
            K_ENTER: exit_game_command,
        }

    def handle(self):
        key = self._screen.getch()
        command = self._command_map.get(key)
        if command:
            if command in [exit_game_command, to_ingame_command, show_popup]:
                command(self._owner)
            else:
                command(self._actor)


# pylint: disable=too-few-public-methods
class GameOverEventHandler(Handler):

    def __init__(self, owner):
        super(GameOverEventHandler, self).__init__(owner)

        self._input_handler = GameOverInputHandler(owner)

    def handle(self):
        self._input_handler.handle()
        # Some other event stuff


class GameOverState(State):

    def __init__(self, owner):
        super(GameOverState, self).__init__(owner)
        self._screen = owner.screen
        self._actor = None

        self._score = "Your score: {0}"

        self._objects = [
            TextWidget(Point(4, 4), "Your soul completely lost this time"),
            TextCallbackWidget(Point(4, 5), self.score_callback),
            MenuItemWidget(Point(10, 10), "I agree"),
            MenuItemWidget(Point(10, 11), "No, I want more")
        ]
        self._objects[2].select()
        self._events = GameOverEventHandler(self)

    def score_callback(self):
        return self._score

    def trigger(self, score):
        """Trigger the state and pass the score info to it.

        :param int score: last player score
        """

        self._score = self._score.format(score)

    def events(self):
        self._events.handle()

    def update(self):
        for o in self._objects:
            o.update()

    def render(self):
        self._screen.erase()
        self._screen.border(0)
        render_objects(self._objects, self._screen)



"""Test xoinvader.collision module."""

import pytest

from xoinvader import collision
from xoinvader.collision import (
    Collider, CollisionManager, CollisionManagerNotFound, TypePair
)
from xoinvader.utils import Point


# pylint: disable=invalid-name,protected-access
def test_collision(mock_state):
    """Generic test for Collider and CollisionManager."""

    class Obj(object):
        def __init__(self, pos):
            self.pos = pos

        def type(self):
            return self.__class__.__name__

    state = mock_state(mock_app=True)

    with pytest.raises(CollisionManagerNotFound):
        Collider(Obj(Point()), [])

    cm = CollisionManager()
    state.collision = cm

    c1 = Collider(
        Obj(Point()), [
            "##",
            "##",
        ])

    c2 = Collider(
        Obj(Point()), [
            "..",
            ".."
        ])

    assert cm.check_collision(c1, c2) is None

    c3 = Collider(
        Obj(Point()), [
            "##",
            "##"
        ])
    assert cm.check_collision(c1, c3)

    c3.obj.pos = Point(1, 0)
    assert cm.check_collision(c1, c3)
    c3.obj.pos = Point(0, 1)
    assert cm.check_collision(c1, c3)
    c3.obj.pos = Point(1, 1)
    assert cm.check_collision(c1, c3)

    c3.obj.pos = Point(0, 2)
    assert cm.check_collision(c1, c3) is None
    c3.obj.pos = Point(2, 0)
    assert cm.check_collision(c1, c3) is None

    c4 = Collider(
        Obj(Point()), [
            "..",
            ".#",
        ])

    assert cm.check_collision(c1, c4)
    c4.obj.pos = Point(1, 0)
    assert cm.check_collision(c1, c4) is None

    assert len(cm._colliders) == 4
    cm.remove(c4)
    assert len(cm._colliders) == 3


def test_type_pair():
    p1 = TypePair("t1", "t2")
    p2 = TypePair("t2", "t1")
    assert p1._pair != p2._pair
    assert p1 != p2
    assert hash(p1) != hash(p2)
    assert str(p1) == "TypePair(t1, t2)"


def test_manager_update(mock_state):

    state = mock_state(mock_app=True)
    cmanager = CollisionManager()
    state.collision = cmanager

    # pylint: disable=too-few-public-methods
    class Ship(object):
        def __init__(self):
            self.health = 10
            self.pos = Point(0, 10)
            self._collider = Collider(
                self, [
                    "..#..",
                    "#####",
                    ".#.#."
                ])

        def type(self):
            return self.__class__.__name__

        @collision.register("Ship", "Rocket")
        def collide(self, other, rect):
            self.health -= other.damage

    class Rocket(object):

        def __init__(self):
            self.damage = 10
            self.pos = Point(0, 0)
            self._collider = Collider(
                self, [
                    "#",
                    "#",
                    "#",
                ])

        def type(self):
            return self.__class__.__name__

    ship = Ship()
    rocket = Rocket()

    cmanager.update()
    assert ship.health == 10

    rocket.pos = Point(0, 8)
    cmanager.update()
    assert ship.health == 10

    rocket.pos = Point(0, 9)
    cmanager.update()
    assert ship.health == 0



"""Game weapon classes."""


from xoinvader.charge import (
    BasicPlasmaCannon, EBasicPlasmaCannon,
    BasicLaserCharge, BasicUnguidedMissile
)
from xoinvader.common import Settings, get_json_config
from xoinvader.render import Renderable
from xoinvader.sound import Mixer
from xoinvader.utils import Point, Surface, Timer


CONFIG = get_json_config(Settings.path.config.weapons)
INFINITE = "infinite"


# TODO: think about composition
# pylint: disable=too-many-instance-attributes,too-many-arguments
class Weapon(Renderable):
    """Main weapon class that implements main methods and behaviour.

    Callbacks for rendering: get_render_data(), remove_obsolete(pos).
    """

    allowed_charges = []
    default_charge = None

    def __init__(self, ammo, max_ammo, cooldown):
        self._type = self.__class__.__name__
        self._image = Surface([" "])  # stub for now, to render nothing
        self._current_charge = self.default_charge
        self._ammo = ammo
        self._max_ammo = max_ammo
        self._cooldown = cooldown
        self._current_cooldown = self._cooldown

        self.ready = True
        self._timer = Timer(self._cooldown, self._reload)
        self._pos = Point()  # stub for now, to render nothing as weapon
        self._loud = True

        Mixer().register(self._type, Settings.path.sound.weapon[self._type])

    def _reload(self):
        """Calls by timer when weapon is ready to fire."""

        # TODO: Play sound
        self.ready = True
        self._timer.stop()
        self._timer.reset()

    def make_shot(self, pos):
        """Check load and ammo, perform shot if ready.

        :param xoinvader.utils.Point pos: position of fire
        :raise ValueError: when no ammo
        """

        if not self.ready:
            return

        if self._ammo == INFINITE:
            self._current_charge(pos)  # pylint: disable=not-callable
        elif self._ammo > 0:
            self._current_charge(pos)  # pylint: disable=not-callable
            self._ammo -= 1

        if self._ammo == 0:
            # No need to measure time.
            # TODO: don't raise exception:
            #       if no ammo - notify player once and
            #       repeat on fire key pressed again.
            self._timer.stop()
            raise ValueError("No ammo!")

        self.ready = False
        self._timer.start()
        if self._loud:
            Mixer().play(self._type)

    def get_render_data(self):
        return ([self._pos], self._image.get_image())

    @property
    def ammo(self):
        """Return actual ammo."""
        return 999 if self._ammo == INFINITE else self._ammo

    @property
    def max_ammo(self):
        """Return maximal ammo."""
        return 999 if self._max_ammo == INFINITE else self._max_ammo

    @property
    def type(self):
        """Return weapon type."""
        return self._type

    def load_percentage(self):
        """Return weapon load percentage."""

        if self._ammo and self.ready:
            return 100.0
        return self._timer.get_elapsed() * 100.0 / self._cooldown

    def update(self):
        """Update weapon timer."""

        if self.ready:
            return

        self._timer.update()


class Blaster(Weapon):
    """Basic player's weapon. Low damage, fast cooldown."""

    allowed_charges = [BasicPlasmaCannon]
    default_charge = BasicPlasmaCannon

    def __init__(self):
        super(Blaster, self).__init__(**CONFIG[self.__class__.__name__])


class EBlaster(Weapon):
    """Basic enemy blaster. Almost identical to Blaster."""

    allowed_charges = [EBasicPlasmaCannon]
    default_charge = EBasicPlasmaCannon

    def __init__(self):
        super(EBlaster, self).__init__(**CONFIG[self.__class__.__name__])


class Laser(Weapon):
    """Basic player's laser. Medium damage, medium cooldown."""

    allowed_charges = [BasicLaserCharge]
    default_charge = BasicLaserCharge

    def __init__(self):
        super(Laser, self).__init__(**CONFIG[self.__class__.__name__])


class UM(Weapon):
    """Player's unguided missile. High damage, slow cooldown."""

    allowed_charges = [BasicUnguidedMissile]
    default_charge = BasicUnguidedMissile

    def __init__(self):
        super(UM, self).__init__(**CONFIG[self.__class__.__name__])



"""Collision detection system and component module."""

import functools
import logging
import weakref

from xoinvader import application
from xoinvader.utils import Point


LOG = logging.getLogger(__name__)


COLLISIONS = {}
"""Global mapping TypePair <=> [callable]."""


class CollisionManagerNotFound(Exception):
    """Raises on try to register collider without instantiated manager."""

    def __init__(self):
        super(CollisionManagerNotFound, self).__init__(
            "You can't use Collider objects without "
            "CollisionManager. Please create it first.")


class TypePair(object):
    """Class for hashable ordered string pairs.

    Used as collision dictionary keys, containing pair of collider types.
    Not commutative, TypePair(a, b) != TypePair(b, a)
    It's needed to store and get exact handler as it was registered.

    :param str first: first collider type
    :param str second: second collider type
    """

    def __init__(self, first, second):
        self._first = first
        self._second = second
        self._pair = first + '_' + second

    @property
    def first(self):
        """First collider type.

        :getter: yes
        :setter: no
        :type: type
        """
        return self._first

    @property
    def second(self):
        """Second collider type.

        :getter: yes
        :setter: no
        :type: type
        """
        return self._second

    def __eq__(self, other):
        return self._pair == other._pair  # pylint: disable=protected-access

    def __hash__(self):
        return hash(self._pair)

    def __str__(self):
        return "TypePair({0}, {1})".format(self._first, self._second)


def register(left, right):
    """Collision handler registration decorator.

    .. Note:: Argument order matters! Handler must belong to first object.

    :param str left: first collidable object
    :param str right: right collidable object
    """

    def decorator(handler):
        COLLISIONS.setdefault(TypePair(left, right), []).append(handler)

        @functools.wraps(handler)
        def handle(*args, **kwargs):
            return handler(*args, **kwargs)

        return handle

    return decorator


class CollisionManager(object):
    """Class for collision detection between known components.

    To process collisions, first update the positions of all objects of
    interest, then call `update` method. It will traverse all registered
    collisions (between pairs of types) and call appropriate handlers in order
    they were registered for the two types of colliding objects.

    If you just want to check, if two Colliders collide, call `check_collision`
    on them.
    """

    # Marker of solid matter inside collider physics map
    __solid_matter__ = "#"

    def __init__(self):
        self._colliders = weakref.WeakSet()
        self._collisions = COLLISIONS

    def add(self, collider):
        """Add collider.

        :param :class:`xoinvader.collision.Collider` collider:
        """

        LOG.debug("Adding collider %s\n pos: %s", collider, collider.pos)
        self._colliders.add(collider)

    def remove(self, collider):
        """Remove collider.

        :param :class:`xoinvader.collision.Collider` collider:
        """

        LOG.debug("Removing collider %s\n pos %s", collider, collider.pos)
        self._colliders.remove(collider)

    # pylint: disable=too-many-nested-blocks
    def update(self):
        """Detect and process all collisions."""

        for pair in self._collisions:
            colliders_type_1 = [
                item for item in self._colliders
                if item.col_type == pair.first]
            colliders_type_2 = [
                item for item in self._colliders
                if item.col_type == pair.second]
            for collider_1 in colliders_type_1:
                for collider_2 in colliders_type_2:
                    collision_rect = self.check_collision(
                        collider_1, collider_2)
                    if collision_rect:
                        for callback in self._collisions[pair]:
                            callback(collider_1.obj, collider_2.obj,
                                     collision_rect)

    # pylint: disable=too-many-locals
    @staticmethod
    def check_collision(col_1, col_2):
        """Check collisions between two colliders.

        Returns `None` if no collision occured, or returns rectangle of
        overlapping region between collider maps.

        :param col1: first collider
        :type col1: :class:`Collider`
        :param col2: second collider
        :type col2: :class:`Collider`
        :rtype: tuple of two :class:`Point`
        """

        width_1 = max(map(len, col_1.phys_map))
        height_1 = len(col_1.phys_map)
        topleft_1 = col_1.pos
        botright_1 = topleft_1 + Point(width_1, height_1)

        width_2 = max(map(len, col_2.phys_map))
        height_2 = len(col_2.phys_map)
        topleft_2 = col_2.pos
        botright_2 = topleft_2 + Point(width_2, height_2)
        if (
                topleft_1.x >= botright_2.x or topleft_1.y >= botright_2.y or
                botright_1.x <= topleft_2.x or botright_1.y <= topleft_2.y
        ):
            # Definelty not overlapping
            return
        # Now find where exactopleft_y overlapping occured
        topleft_overlap = Point(
            max(topleft_1.x, topleft_2.x),
            max(topleft_1.y, topleft_2.y))
        botright_overlap = Point(
            min(botright_1.x, botright_2.x),
            min(botright_1.y, botright_2.y))
        # Now find if they actually collided
        # first, calculate offsets
        overlap_1 = Point()
        overlap_2 = Point()
        overlap_1.x = abs(topleft_1.x - topleft_overlap.x)
        overlap_1.y = abs(topleft_1.y - topleft_overlap.y)
        overlap_2.x = abs(topleft_2.x - topleft_overlap.x)
        overlap_2.y = abs(topleft_2.y - topleft_overlap.y)
        # iterate over overlapping region
        # and search for collision
        for i in range(botright_overlap.x - topleft_overlap.x):
            for j in range(botright_overlap.y - topleft_overlap.y):
                # TODO: check length of current y-level string
                # it might be not enough to contain i + ox1/2 element
                if (
                        col_1.phys_map[j + overlap_1.y][i + overlap_1.x] ==
                        col_2.phys_map[j + overlap_2.y][i + overlap_2.x] ==
                        CollisionManager.__solid_matter__
                ):
                    return (topleft_overlap, botright_overlap)


class Collider(object):
    """Collider component class.

    When added to object, enables it to participate in coliision processing
    system: i.e. to be able to detect and process collisions between the object
    and other ones.

    .. Attention:: CollisionManager must be created first and must be
                   accessible via State.

    :param object obj: GameObject to which the collider is linked
    :param list phys_map: list of strings representing collider physical
    geometry. All strings must be of equal length. Class member
    __solid_matter__ of CollisionManager represents solid geometry, all other
    chars are treated as void space and may be any.
    """

    def __init__(self, obj, phys_map):
        self._obj = obj
        self._col_type = self._obj.type()
        self._phys_map = phys_map

        # TODO: move collision to State.systems
        try:
            application.get_current().state.collision.add(self)
        except AttributeError:
            raise CollisionManagerNotFound()

    @property
    def phys_map(self):
        """Collider physical geometry.

        :getter: yes
        :setter: no
        :type: list
        """
        return self._phys_map

    @property
    def col_type(self):
        """Collider type name.

        :getter: yes
        :setter: no
        :type: str
        """
        return self._col_type

    @property
    def pos(self):
        """Collider's left top position.

        :getter: yes
        :setter: no
        :type: :class:`Point`
        """
        return self._obj.pos[int]

    @property
    def obj(self):
        return self._obj



"""Test xoinvader.weapon module."""

import pytest

import xoinvader
from xoinvader import weapon


# pylint: disable=missing-docstring
@pytest.mark.skip
def test_weapon(monkeypatch):
    monkeypatch.setattr(weapon, "CONFIG", {})
    monkeypatch.setattr(weapon, "Mixer", xoinvader.sound.get_mixer)

    # pytest.raises(KeyError, lambda:
    weapon.Weapon(ammo=10, max_ammo=10, cooldown=0.5)



"""Pytest configuration."""

import pytest

from xoinvader import application
from xoinvader import state


@pytest.fixture
def mock_application(request):

    def inner():
        return MockedApplication()

    request.addfinalizer(MockedApplication._finalize)
    return inner


@pytest.fixture
def mock_state(request):

    def inner(mock_app=False):
        if mock_app:
            MockedApplication()
        application.get_current().register_state(MockedState)
        return application.get_current().state

    def stop():
        application.get_current().deregister_state(MockedState.__name__)
        if not application.get_current().states:
            MockedApplication._finalize()

    request.addfinalizer(stop)

    return inner


class MockedApplication(application.Application):

    @staticmethod
    def _finalize():
        app = application.get_current()
        app.stop()


class MockedState(state.State):
    pass



"""Vector data structure."""

from functools import wraps
from math import sqrt
from typing import Any, TypeVar, Union, Tuple, Callable

import numpy as np


Numeric = TypeVar('Numeric', int, float)
#ForwardRef = TypeVar('ForwardRef', str)
OperandType = Union[str, type]#Union[ForwardRef, type]


def hypot(x: Numeric, y: Numeric, z: Numeric = 0) -> Numeric:
    return sqrt(x * x + y * y + z * z)


class OperandTypeError(ValueError):
    """Raises when the operand has unsupported type for the operation."""

    def __init__(self, value: Any, opname: str, optype: type) -> None:
        super().__init__(
         f'Operand {value} has unsupported type {type(value).__name__}, must be {optype.__name__}'
        )


def singlecontract(opname: str, optype: Union[Tuple[OperandType], OperandType]) -> Callable:
    """Simple runtime type checker for first function argument."""

    def wrapper(func):
        @wraps(func)
        def guard(value, *args, **kwargs):

            optypes = (optype,) if not isinstance(optype, tuple) else optype
            resolved = []
            for optype in optypes:
                if isinstance(optype, str):
                    resolved.append(_parse_type(optype))
                else:
                    resolved.append(optype)

            if not isinstance(value, tuple(resolved)):
                raise OperandTypeError(value, opname, optype)

            return func(value, *args, **kwargs)

        return guard

    return wrapper


def _parse_type(type_literal: str) -> type:
    try:
        return eval(compile(type_literal, '<string>', 'eval'))
    except SyntaxError:
        raise SyntaxError(f'Forward reference must be an expression, got {type_literal!r}')


class Vec2:

    __slots__ = ('x', 'y')

    def __init__(self, x: Numeric = 0, y: Numeric = 0) -> None:
        self.x = x
        self.y = y

    def __repr__(self) -> str:
        return f'Vec2(x={self.x}, y={self.y})'

    def __abs__(self) -> float:
        return hypot(self.x, self.y)

    def __add__(self, other: 'Vec2') -> 'Vec2':
        pass

    def __sub__(self, other):
        pass

    def __mul__(self, other):
        if isinstance(other, Vec2):
            pass
        elif isinstance(other, (int, float)):
            pass
        else:
            raise ValueError(f'Type {0} is unsupported for {1}')

    def __truediv__(self, other):
        pass

    def __eq__(self, other):
        if isinstance(other, Vec2):
            return self.x == other.x and self.y == other.y

        raise ValueError()

class Vec3:

    __slots__ = ('x', 'y', 'z')

    def __init__(self, x=0, y=0, z=0):
        self.x = x
        self.y = y
        self.z = z

    @property
    def opposite(self, other):
        pass

    def __repr__(self):
        return f'Vec3(x={self.x}, y={self.y}, z={self.z})'

    def __abs__(self):
        return hypot(self.x, self.y, self.z)

    @singlecontract('add', 'Vec3')
    def __add__(self, other):
        """Vec3(1, 2, 3) + Vec3(4, 5, 6) = Vec3(5, 7, 9)"""

        return Vec3(x=self.x + other.x, y=self.y + other.y, z=self.z + other.z)

    @singlecontract('sub', 'Vec3')
    def __sub__(self, other):
        """Vec3(3, 3, 3) - Vec3(2, 2, 2) = Vec3(1, 1, 1)"""

        return self.__add__(-other)

    @singlecontract('mul', (int, float, 'Vec3'))
    def __mul__(self, other):
        """-->
        * Vec3(3, 3, 3) * 3 = Vec3(9, 9, 9)
        * Vec3(3, 3, 3) * 
        """

        raise NotImplementedError

    def __truediv__(self, other):
        pass

    @singlecontract('eq', 'Vec3')
    def __eq__(self, other):

        return self.x == other.x and self.y == other.y and self.z == other.z



"""Level background."""

from xoinvader.common import Settings
from xoinvader.render import Renderable
from xoinvader.utils import Point, Surface


CHUNK_MAGIC = "~chunk~"
"""Background file format chunk marker."""


class Chunk(object):
    """Class for storing background chunks.

    :param str name: name of the chunk
    :param list lines: list of all chunk lines
    """

    def __init__(self, name):
        self._name = name
        self._lines = []

    @property
    def name(self):
        """Chunk name.

        :getter: yes
        :setter: no
        :type: str
        """
        return self._name

    @property
    def lines(self):
        """All chunk lines.

        :getter: yes
        :setter: no
        :type: list
        """
        return self._lines

    def add_line(self, line):
        """Add line to `lines` list of the chunk.

        :param str line: line to add
        """
        self._lines.append(line)

    def __getitem__(self, index):
        return self._lines[index]

    def __setitem__(self, index, item):
        self._lines[index] = item

    def __len__(self):
        return len(self._lines)


def load_chunks(filename, trim_width=None):
    """Load chunks from file.

    File format:
    ```
    -~chunk~- <chunk_name>
    <data>
    ...
    ```

    `~chunk~` is magic string, and chunk spans until next magic line, or
    until end of file.

    Completely blank lines are ignored.
    If `trim_width` is provided, all lines are truncated to this length.

    :param str filename: path to file to load chunks from
    :param int trim_width: width to truncate long lines to
    :return: list of loaded chunks
    :rtype: list
    """

    chunks = []
    current_chunk = None
    names = []
    with open(filename, "r") as f:
        for line in f:
            line = line.rstrip("\n")
            if not line:
                continue
            elif line.startswith(CHUNK_MAGIC):
                name_split = line.split()
                if len(name_split) < 2:
                    raise ValueError("Chunks with empty names are not allowed")
                name = name_split[1].strip()
                if name in names:
                    raise ValueError("Name {0} already defined in background "
                                     "file {1}".format(name, filename))
                current_chunk = Chunk(name)
                chunks.append(current_chunk)
                names.append(name)
                continue
            elif current_chunk is None:
                continue
            # ok, now we got both chunk and something to store in it
            line_add = line[:trim_width] if trim_width else line
            current_chunk.add_line(line_add)
    if not chunks:
        raise ValueError("File {0} does not contain "
                         "any chunks".format(filename))
    return chunks


# pylint: disable=too-many-instance-attributes
class Background(Renderable):
    """Class for storing and managing level background.

    Manages background rendering and changing in time.

    :param bool loop: flag if current chunk should loop after it had ended
    :param bool loop_all: flag if whole background should repeat from beginning
    :param float speed: speed of background advance. Can be negative to
    represent moving backwards
    :param list chunks: list of background chunks
    :param list background: list of background lines. You may change it at any
    time. This list directly converts into Surface which then goes to the
    renderer
    """

    render_priority = -1000  # TODO: render-priority

    def __init__(self, filename=None, speed=0, loop=False, loop_all=False):
        self._speed = speed
        self._loop = loop
        self._loop_all = loop_all
        self._chunks = []
        self._background = []

        # pylint: disable=invalid-name
        self._w = Settings.layout.field.edge.x
        self._h = Settings.layout.field.edge.y
        self._background_surface = None
        self._current_chunk = None
        self._current_chunk_num = 0  # position of chunk in chunk list
        self._chunk_line = 0  # position in current chunk
        self._ticks_since_last_update = 0

        if filename:
            self.load_file(filename)

    @property
    def loop(self):
        """Loop current chunk.

        :getter: yes
        :setter: yes
        :type: bool
        """
        return self._loop

    @loop.setter
    def loop(self, value):
        """Setter."""
        self._loop = value

    @property
    def loop_all(self):
        """Loop all chunks.

        :getter: yes
        :setter: yes
        :type: bool
        """
        return self._loop_all

    @loop_all.setter
    def loop_all(self, value):
        """Setter."""
        self._loop_all = value

    @property
    def speed(self):
        """Speed of background advance.

        :getter: yes
        :setter: yes
        :type: int
        """
        return self._speed

    @speed.setter
    def speed(self, value):
        """Setter."""
        self._speed = value

    @property
    def chunks(self):
        """Background chunks.

        :getter: yes
        :setter: yes
        :type: list
        """
        return self._chunks

    @chunks.setter
    def chunks(self, value):
        """Setter."""
        self._chunks = value

    @property
    def background(self):
        """Background lines.

        :getter: yes
        :setter: yes
        :type: list
        """
        return self._background

    @background.setter
    def background(self, value):
        """Setter."""
        self._background = value

    def clear(self):
        """Clear all background.

        Doesn't prevent subsequent updating, if speed is not 0.
        """

        self._background = [" " * self._w] * self._h

    def load_file(self, filename):
        """Load background from file.

        Calls `load_chunks` function int trim mode, and stores return value in
        the `chunks` field.

        :param str filename:
        """

        self._chunks = load_chunks(filename, self._w)

    def _fill(self):
        """Fills the whole background.

        Takes first chunk and fills background with it's contents. If there's
        not enough lines, switches to next chunk, etc. I.e, we just advance
        background momentary, and update all pointers accordingly.
        """

        self._background = []
        for _ in range(self._h):
            self._background.append(self._advance_chunk(1))

    def start(self, filled=False):
        """Start background from the beginning.

        If `filled` parameter is provided, fills the background starting with
        first chunk, using `_fill` method.

        :param bool filled: if true, perform initial background fill
        """

        self._ticks_since_last_update = 0
        self._current_chunk_num = 0
        self._current_chunk = self._chunks[0]
        self._chunk_line = 0
        if filled:
            self._fill()
        else:
            self.clear()
        self.update_surface()

    def _advance_chunk(self, advance):
        """Return current chunk line and update pointers.

        Performs checks if current chunk number and current line position are
        valid, updates them, if they're not, returns current line and then
        updates current line position. I.e., performs boundary checks, updates
        if necessary and then updates current line position.

        :return: current line in chunk sequence
        :rtype: string
        """

        if not self._current_chunk:
            return " " * self._w

        # start checks
        if (self._chunk_line < len(self._current_chunk) and
                self._chunk_line >= 0):
            line = self._current_chunk[self._chunk_line]
            self._chunk_line += advance
            return line

        # chunk ended! now we decide what to do
        if self._loop:
            if self._chunk_line < 0:
                self._chunk_line = len(self._current_chunk) - 1
            else:
                self._chunk_line = 0
            line = self._current_chunk[self._chunk_line]
            self._chunk_line += advance
            return line

        # next chunk or stop advancing
        self._current_chunk_num += advance
        if (self._current_chunk_num < len(self._chunks) and
                self._current_chunk_num >= 0):
            # okay, just set new chunk
            self._current_chunk = self._chunks[self._current_chunk_num]
            self._chunk_line = 0
            line = self._current_chunk[self._chunk_line]
            self._chunk_line += advance
            return line

        if self._loop_all:
            if self._current_chunk_num < 0:
                self._current_chunk_num = len(self._chunks) - 1
                self._current_chunk = self._chunks[-1]
                self._chunk_line = len(self._current_chunk) - 1
            else:
                self._current_chunk_num = 0
                self._current_chunk = self._chunks[0]
                self._chunk_line = 0
            line = self._current_chunk[self._chunk_line]
            self._chunk_line += advance
            return line

        # looks like we don't have any more chunks, and we're not looping
        # thus return void line
        self._current_chunk = None
        return " " * self._w

    def get_render_data(self):
        return [Point(0, 0)], self._background_surface.get_image()

    def update_surface(self):
        """Regenerate background surface.

        Intended to be call after all update operations on the `background`
        member list so updated info will immediately come into effect.
        """

        self._background_surface = Surface(self._background)

    def update(self):
        """Update background.

        Checks if time to update has come, and calls `_advance_chunk` method
        with appropriate parameter. Calls `update_surface` at the end.

        """

        if self._speed == 0:
            return

        self._ticks_since_last_update += abs(self._speed)
        if self._ticks_since_last_update < 60:  # TODO FIXME: hardcode wololo
            return

        self._ticks_since_last_update = 0

        advance = 1 if self._speed > 0 else -1
        new_line = self._advance_chunk(advance)
        if advance > 0:
            self._background.pop()
            self._background.insert(0, new_line)
        else:
            self._background.pop(0)
            self._background.append(new_line)
        self.update_surface()



"""Game weapon charge classes.

Attention: Next statement related only to GameObject entities.
           There're situations where two GameObject modules need to import
           each other. In such cases do it via importing `xoinvader.registry`
           in functions (not at import time).
"""


import curses
import logging
import re

from xoinvader import application
from xoinvader.collision import Collider
from xoinvader.common import Settings, get_json_config
from xoinvader.render import Renderable
from xoinvader.utils import Point, Surface


CONFIG = get_json_config(Settings.path.config.charges)
LOG = logging.getLogger(__name__)


# TODO: [components]: move some variables to default GameObject components
# pylint: disable=too-many-arguments,too-many-instance-attributes
class WeaponCharge(Renderable):
    """Weapon charge representation.

    Separate renderable object that updates and renders as others
    in main loop. Must register/deregister itself via state's
    add/remove methods.
    """

    def __init__(self, pos, image, damage=0, radius=0, dx=0, dy=0):
        self._pos = pos
        self._image = image
        self._type = self.__class__.__name__
        application.get_current().state.add(self)

        self._damage = damage
        self._radius = radius
        # TODO: [object-movement]
        # Generalize update to support dx
        self._dx = dx
        self._dy = dy

        # Set to True means that object in destroying phase
        # and will ignore remove_obsolete signal
        self._destroy = False

        # Common collider for any shape charge
        # TODO FIXME: objects want know, what '#' means!
        #             make solid marker accessible to them
        #             and/or provide 'Collider.from_array(self._image.raw)'
        self._collider = Collider(
            self, [
                re.sub(r"[^\ ]", "#", row)
                for row in self._image.raw
            ])

    @property
    def pos(self):
        return self._pos

    @property
    def damage(self):
        return self._damage

    def get_render_data(self):
        return ([self._pos], self._image.get_image())

    def remove_obsolete(self, pos):
        border = Settings.layout.field.border
        pos = self._pos[int]
        if (
                pos.x > self._image.width + border.x or
                pos.x + self._image.width < 0 or
                int(pos.y) > self._image.height + border.y or
                int(pos.y) + self._image.height < 0
        ) and not self._destroy:
            application.get_current().state.remove(self)
            self._destroy = True

    def update(self):
        """Update coords."""

        self._pos += Point(self._dx, self._dy)

    def destroy(self):
        """Self-destroying routine."""

        if not self._destroy:
            self._destroy = True
            application.get_current().state.collision.remove(self._collider)
            application.get_current().state.remove(self)


# TODO: Implement hitscan behaviour
# pylint: disable=useless-super-delegation
class Hitscan(WeaponCharge):
    """Hitscan weapon charge hits target immediately.

    Can pierce enemies?
    """

    def __init__(self, pos, image, **kwargs):
        super(Hitscan, self).__init__(pos, image, **kwargs)


class Projectile(WeaponCharge):
    """Projectile weapon has generic physics of movement."""

    def __init__(self, pos, image, **kwargs):
        super(Projectile, self).__init__(pos, image, **kwargs)


# TODO: write template documentation for charges and weapons
class BasicPlasmaCannon(Projectile):
    """Small damage, no radius."""

    def __init__(self, pos):
        super(BasicPlasmaCannon, self).__init__(
            pos, Surface(["^"], style=[[curses.A_BOLD]]),
            **CONFIG[self.__class__.__name__])


class EBasicPlasmaCannon(Projectile):
    """Enemy plasma cannon."""

    def __init__(self, pos):
        super(EBasicPlasmaCannon, self).__init__(
            pos, Surface([":"], style=[[curses.A_BOLD]]),
            **CONFIG[self.__class__.__name__])


class BasicLaserCharge(Projectile):
    """Laser. Quite fast but cannot pierce enemies."""

    def __init__(self, pos):
        super(BasicLaserCharge, self).__init__(
            pos, Surface(["|"], style=[[curses.A_BOLD]]),
            **CONFIG[self.__class__.__name__])


class BasicUnguidedMissile(Projectile):
    """Unguided missile with medium damage and small radius."""

    def __init__(self, pos):
        super(BasicUnguidedMissile, self).__init__(
            pos, Surface([
                "^",
                "|",
                "*"
            ], style=[[curses.A_BOLD] for _ in range(3)]),
            **CONFIG[self.__class__.__name__])



"""Various useful tools."""

import copy
import datetime
import logging
import time
try:
    time.perf_counter
except AttributeError:
    time.perf_counter = time.time


LOG_FORMAT = (
    "[%(asctime)s] %(levelname)-8s %(name)s[%(funcName)s]:%(lineno)s:  "
    "%(message)s"
)
"""Log message format string."""

TIME_FORMAT = "%H:%M:%S,%03d"
"""Log time format string."""

DATE_FORMAT = "%Y-%m-%d %a"
"""Initial log entry date format string."""


def setup_logger(name, debug=False, msgfmt=None, timefmt=None):
    """Setup logger with linked log file.

    Do not use it for getting logger, call this once on init,
    then use logging.getLogger(__name__) for getting actual logger.

    :param str name: logger relative name
    :param bool debug: debug mode
    :param str msgfmt: message format
    :param str timefmt: time format

    :return: prepared logger instance
    :rtype: `logging.Logger`
    """

    logger = logging.getLogger(name)
    logger.propagate = False
    level = logging.DEBUG if debug else logging.INFO
    logger.setLevel(level)
    handler = logging.FileHandler("{0}.log".format(name))
    handler.setLevel(level)
    formatter = logging.Formatter(msgfmt or LOG_FORMAT, timefmt or TIME_FORMAT)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    date = datetime.date.today().strftime(DATE_FORMAT)
    logger.info("*** (%s) Initializing XOInvader ***", date)

    return logger


def isclose(left, right, rel_tol=1e-9, abs_tol=0.0):
    """Check if values are approximately equal.

    :param int,float left: first value
    :param int,float right: second value
    :param float rel_tol: relative tolerance (amount of error allowed)
    :param float abs_tol: minimum absolute tolerance level
    """
    approx = max(rel_tol * max(abs(left), abs(right)), abs_tol)
    return abs(left - right) <= approx


class dotdict(dict):  # pylint: disable=invalid-name
    """Container for dot elements access."""

    def __init__(self, *args, **kwargs):
        super(dotdict, self).__init__(*args, **kwargs)
        self.__dict__ = self
        self._wrap_nested()

    def _wrap_nested(self):
        """Wrap nested dicts for deep dot access."""

        for key, value in self.items():
            if isinstance(value, dict):
                self[key] = dotdict(value)

    def fullcopy(self):
        """Return full copy of internal structure as dotdict.

        :return :class:`xoinvader.utils.dotdict`: full copy
        """

        return dotdict(copy.deepcopy(self))


# pylint: disable=too-few-public-methods
class Point(object):
    """3D point representation.

    :param x: x coordinate
    :type x: integer

    :param y: y coordinate
    :type y: integer

    :param z: z coordinate
    :type z: integer

    :return: Point instance
    :rtype: `xoinvader.utils.Point`
    """

    __slots__ = ["x", "y", "z"]

    def __init__(self, x=0, y=0, z=0):
        self.x = x
        self.y = y
        self.z = z

    def __repr__(self):
        return "Point(x={0}, y={1}, z={2})".format(self.x, self.y, self.z)

    __str__ = __repr__

    @staticmethod
    def _value_error(operation, value):
        """Raise ValueError with appropriate message."""

        return ValueError(
            "Wrong type to {0} {1}: {2}".format(operation, type(value), value))

    def __add__(self, other):
        if isinstance(other, (int, float)):
            return Point(
                x=self.x + other,
                y=self.y + other,
                z=self.z + other)

        elif isinstance(other, Point):
            return Point(
                x=self.x + other.x,
                y=self.y + other.y,
                z=self.z + other.z)
        else:
            raise self._value_error("add", other)

    def __sub__(self, other):
        if isinstance(other, (int, float)):
            return self.__add__(-other)
        elif isinstance(other, Point):
            return Point(
                x=self.x - other.x,
                y=self.y - other.y,
                z=self.z - other.z)
        else:
            raise self._value_error("sub", other)

    def __mul__(self, other):
        if isinstance(other, (int, float)):
            return Point(
                x=self.x * other,
                y=self.y * other,
                z=self.z * other)
        # What for point
        else:
            raise self._value_error("mul", other)

    def __truediv__(self, other):
        if isinstance(other, (int, float)):
            return Point(
                x=self.x / other,
                y=self.y / other,
                z=self.z / other)
        else:
            raise self._value_error("div", other)

    __div__ = __truediv__

    def __eq__(self, other):
        if not isinstance(other, Point):
            raise self._value_error("eq", other)

        return self.x == other.x and self.y == other.y and self.z == other.z

    def __getitem__(self, cons):
        """Cast Point to selected type.

        :param type cons: type to cast to

        :return Point: with casted members
        """

        return Point(
            x=cons(self.x),
            y=cons(self.y),
            z=cons(self.z))


class Surface(object):
    """Representation of graphic objects.

    :param image: source for building image
    :type image: [ [char] ]

    .. note:: add building from string

    :param style: source for building styled image
    :type style: [ [integer(curses constant)] ]

    :return: Surface instance
    :rtype: `xoinvader.utils.Surface`
    """

    # Example:
    #    rocket
    #    [
    #     ["^"],     ^
    #     ["|"],     |
    #     ["*"] ]    *

    #    ship
    #    [
    #     [" "," ","O"," "," "],           O
    #     ["<","=","H","=",">"],         <=H=>
    #     [" ","*"," ","*"," "] ]         * *

    __slots__ = ["_image", "_width", "_height", "_style"]

    def __init__(self, image, style=None):
        self._image = image
        self._width = max([len(l) for l in self._image])
        self._height = len(self._image)
        self._style = style

    @property
    def height(self):
        """Height of the surface.

        :getter: yes
        :setter: no
        :type: integer
        """
        return self._height

    @property
    def width(self):
        """Width of the surface.

        :getter: yes
        :setter: no
        :type: integer
        """
        return self._width

    @property
    def raw(self):
        """Raw image content.

        :getter: yes
        :setter: no
        :type: list
        """
        return self._image

    def get_image(self):
        """Image generator. Allows to renderers render Surfaces.

        :return: image generator
        :rtype: generator(`xoinvader.utils.Point`, string, integer)
        """
        for y, row in enumerate(self._image):
            for x, image in enumerate(row):
                yield (
                    Point(x=x, y=y), image,
                    self._style[y][x] if self._style else None
                )


class InfiniteList(list):
    """Infinite list container."""

    def __init__(self, *args, **kwargs):
        super(InfiniteList, self).__init__(*args, **kwargs)
        self._index = 0

    def current(self):
        """Get current element.

        :return: current element
        :rtype: object
        """
        return self[self._index]

    def next(self):
        """Get next element.

        :return: next element
        :rtype: object
        """
        try:
            self._index = (self._index + 1) % len(self)
        except ZeroDivisionError:
            raise IndexError("List is empty.")
        return self[self._index]

    def prev(self):
        """Get previous element.

        :return: previous element
        :rtype: object
        """
        try:
            self._index = (self._index - 1) % len(self)
        except ZeroDivisionError:
            raise IndexError("List is empty.")
        return self[self._index]


class Timer(object):
    """Simple timer, calls callback when time's up. Doesn't have own loop."""

    def __init__(self, end_time, func):
        self._end = float(end_time)
        self._func = func
        self._start = time.perf_counter()
        self._current = self._start
        self._running = False

    def _tick(self):
        """Refresh counter."""
        if self.running:
            self._current = time.perf_counter()

    def _time_is_up(self):
        """return is it time to fire fuction or not.

        :return: time is up
        :rtype: boolean
        """
        return self._current - self._start >= self._end

    def start(self):
        """Start timer."""
        self._running = True
        self._start = time.perf_counter()
        self._current = time.perf_counter()

    def stop(self):
        """Stop timer."""
        self._running = False

    def restart(self):
        """Restart timer."""
        self._start = time.perf_counter()
        self._current = self._start
        self.start()

    def reset(self):
        """Reset timer."""
        self._running = False
        self._start = 0
        self._current = 0

    def update(self):
        """Public method for using in loops."""
        if not self.running:
            return

        # Timer's accuracy depends on owner's loop
        self._tick()
        if self._time_is_up() and self.running:
            self._func()
            self.stop()

    @property
    def running(self):
        """Is timer running or not.

        :getter: yes
        :setter: no
        :type: boolean
        """
        return self._running

    def get_elapsed(self):
        """Elapsed time from start.

        :return: elapsed time
        :rtype: float
        """
        return self._current - self._start

    def get_remaining(self):
        """Remaining time to fire callback.

        :return: remaining time
        :rtype: float
        """
        return self._end - self.get_elapsed()

    def fire_function(self):
        """Call stored callback."""
        self._func()


class Singleton(type):
    """Singleton metaclass."""

    _instances = {}

    def __call__(cls, *args, **kwds):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwds)
        return cls._instances[cls]



"""Test xoinvader.handlers module."""

import pytest

from xoinvader.handlers import Command
from xoinvader.handlers import Handler


# pylint: disable=missing-docstring,too-few-public-methods
class OwnerMock:
    screen = "screen"
    actor = "actor"


def test_command():
    command = Command()
    with pytest.raises(NotImplementedError):
        command.execute("actor")


def test_handler():
    handler = Handler(OwnerMock())
    with pytest.raises(NotImplementedError):
        handler.handle()



"""State for testing on pygame-sdl driver."""

import pygame
import pygame.locals

from pygame.locals import K_a, K_d, K_ESCAPE

from xoinvader.ship import TestShip
from xoinvader.handlers import Handler
from xoinvader.state import State
from xoinvader.common import Settings


# pylint: disable=all
class TestStateInputHandler(Handler):

    def __init__(self, owner):
        super(self.__class__, self).__init__(owner)

        self._commands = {
            pygame.KEYDOWN: {
                K_a: self._actor.move_left,
                K_d: self._actor.move_right,
                K_ESCAPE: self._owner._owner.stop},
            pygame.KEYUP: {}}

    def handle(self, event):
        cmd = self._commands.get(event.type).get(event.key)
        if cmd:
            cmd()


class TestStateEventHandler(Handler):

    def __init__(self, owner):
        super(self.__class__, self).__init__(owner)

        self._input_handler = TestStateInputHandler(owner)

    def handle(self):
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                self._owner._owner.stop()
            elif event.type in [pygame.KEYUP, pygame.KEYDOWN]:
                self._input_handler.handle(event)
        # Some other event logic


class TestState(State):

    def __init__(self, owner):
        super(TestState, self).__init__(owner)
        self._objects = []
        self._screen = self._owner.screen

        self._actor = TestShip(Settings.layout.field.player,
                               Settings.layout.field.edge, Settings)

        self._objects.append(self._actor)
        self._events = TestStateEventHandler(self)

    def events(self):
        self._events.handle()

    def update(self):
        for obj in self._objects:
            obj.update()

    def render(self):
        self._screen.fill(pygame.Color(0, 0, 0, 1))
#        self._screen.blit(self._background, (0, 0))
        for obj in self._objects:
            obj.render(self._screen)

        pygame.display.update()



"""NCurses application class."""

import xoinvader.curses_utils
from xoinvader.common import Settings
from xoinvader.application import Application


class CursesApplication(Application):
    """Curses powered application class, default fallback backend.

    :param dict startup_args: arguments for passing to game settings
    """

    def __init__(self):
        super(CursesApplication, self).__init__()

        self._screen = xoinvader.curses_utils.create_window(
            ncols=Settings.layout.field.border.x,
            nlines=Settings.layout.field.border.y)

        self._clock = xoinvader.curses_utils.get_clock()

    def set_caption(self, caption, icontitle=None):
        """Set window caption.

        :param caption: window caption
        :type caption: str

        :param icontitle: short title
        :type icontitle: str
        """
        # Maybe we can update text via terminal API
        pass

    def stop(self):
        self._ioloop.add_callback(
            lambda: xoinvader.curses_utils.deinit_curses(self._screen))
        super(CursesApplication, self).stop()



"""Curses helper module."""

import time
import curses

from six import add_metaclass

from xoinvader.common import Settings
from xoinvader.utils import Singleton


# pylint: disable=too-few-public-methods
@add_metaclass(Singleton)
class _Color(object):
    """Curses color mapping."""

    def __init__(self):
        self._color_names = [
            # User interface colors
            "ui_norm",
            "ui_yellow",
            # Damage panel colors
            "dp_blank",
            "dp_ok",
            "dp_middle",
            "dp_critical",
            "sh_ok",
            "sh_mid",
            # Weapon's gauge colors
            "blaster",
            "laser",
            "um",
        ]
        self._color_map = dict(
            zip(self._color_names, range(1, len(self._color_names) + 1))
        )

    @property
    def color_names(self):
        """Color names.

        :getter: yes
        :setter: no
        :type: list
        """
        return self._color_names

    def __getattr__(self, name):
        return self._color_map[name]


Color = _Color()  # pylint: disable=invalid-name


# TODO: rewrite this shit
@add_metaclass(Singleton)
class Style(object):
    """Container for style mappings."""

    def __init__(self):
        self._style = dict(
            gui={},
            obj={},
        )

    def init_styles(self, curses_module):
        """Initialize styles.

        :param module curses_module: curses module to initialize pairs
        """

        if Settings.system.no_color:
            for key in Color.color_names:
                self.gui[key] = None

        cpair = curses_module.color_pair

        self.gui["normal"] = cpair(Color.ui_norm) | curses.A_BOLD
        self.gui["yellow"] = cpair(Color.ui_yellow) | curses.A_BOLD
        self.gui["dp_blank"] = cpair(Color.dp_blank) | curses.A_BOLD
        self.gui["dp_ok"] = cpair(Color.dp_ok) | curses.A_BOLD
        self.gui["dp_middle"] = cpair(Color.dp_middle) | curses.A_BOLD
        self.gui["dp_critical"] = cpair(Color.dp_critical) | curses.A_BOLD
        self.gui["sh_ok"] = cpair(Color.sh_ok) | curses.A_BOLD
        self.gui["sh_mid"] = cpair(Color.sh_mid) | curses.A_BOLD

    def __getattr__(self, name):
        return self._style[name]


# TODO: refactor
def get_styles():
    """Return Style object."""
    return Style()


def init_curses_pairs(curses_module):
    """Init curses color pairs.

    :param module curses_module: curses module for pair initialization.
    """

    init_pair = curses_module.init_pair

    # User interface
    init_pair(Color.ui_norm, curses.COLOR_WHITE, curses.COLOR_BLACK)
    init_pair(Color.ui_yellow, curses.COLOR_YELLOW, curses.COLOR_BLACK)

    # Damage panel
    init_pair(Color.dp_blank, curses.COLOR_BLACK, curses.COLOR_BLACK)
    init_pair(Color.dp_ok, curses.COLOR_GREEN, curses.COLOR_BLACK)
    init_pair(Color.dp_middle, curses.COLOR_YELLOW, curses.COLOR_BLACK)
    init_pair(Color.dp_critical, curses.COLOR_RED, curses.COLOR_BLACK)
    init_pair(Color.sh_ok, curses.COLOR_BLUE, curses.COLOR_BLACK)
    init_pair(Color.sh_mid, curses.COLOR_CYAN, curses.COLOR_BLACK)

    # Weapons
    init_pair(Color.blaster, curses.COLOR_GREEN, curses.COLOR_BLACK)
    init_pair(Color.laser, curses.COLOR_BLACK, curses.COLOR_RED)
    init_pair(Color.um, curses.COLOR_MAGENTA, curses.COLOR_BLACK)


def create_window(ncols, nlines, begin_x=0, begin_y=0):
    """Initialize curses, colors, make and return window.

    :param ncols: number of columns
    :type ncols: integer

    :param nlines: number of lines
    :type nlines: integer

    :param begin_x: offset by x
    :type begin_x: integer

    :param begin_y: offset by y
    :type begin_y: integer

    :return: initialized curses window
    :rtype: `curses.Window`
    """
    curses.initscr()
    curses.start_color()

    if not Settings.system.no_color:
        init_curses_pairs(curses)

    # XXX: is this must be the first call of Style?
    Style().init_styles(curses)

    screen = curses.newwin(nlines, ncols, begin_x, begin_y)
    screen.keypad(0)
    screen.nodelay(1)
    curses.noecho()
    curses.cbreak()
    curses.curs_set(0)
    return screen


def deinit_curses(screen):
    """Destroy window, deinit curses, make console changes back.

    :param screen: main screen
    :type screen: `curses.Window`
    """
    screen.nodelay(0)
    screen.keypad(0)
    curses.nocbreak()
    curses.echo()
    curses.curs_set(1)
    curses.endwin()


class Clock():
    """Object that helps to track time."""

    def __init__(self):
        self._mspf = 0
        self._previous_tick = 0
        self._current_tick = int(time.perf_counter())
        self._tick_time = 0

    def tick(self, framerate=0.0):
        """Update the clock.

        :param float framerate: expected FPS
        """

        self._previous_tick = self._current_tick
        self._current_tick = int(time.perf_counter())
        delta = self._current_tick - self._previous_tick

        self._mspf = int(1.0 / framerate * 1000.0) if framerate else delta

        time_s = (self._mspf - delta) / 1000.0
        time.sleep(time_s)
        self._tick_time = int(time.perf_counter()) - self._previous_tick
        return self._tick_time

    def get_fps(self):
        """Compute the clock framerate.

        :return: FPS
        :rtype: float
        """
        pass

    def get_time(self):
        """Time used in previous tick.

        :return: tick duration in milliseconds
        :rtype: int
        """
        return self._tick_time


def get_clock():
    """Helper for unification with other backends."""

    return Clock()



"""Enemy and player ships."""

import logging

from xoinvader import application
from xoinvader import collision
from xoinvader.animation import AnimationManager
from xoinvader.entity import Entity
from xoinvader.sound import Mixer
from xoinvader.render import Renderable
from xoinvader.weapon import Blaster, Laser, UM, EBlaster
from xoinvader.utils import Point, Surface, InfiniteList
from xoinvader.collision import Collider
from xoinvader.common import Settings, get_json_config


LOG = logging.getLogger(__name__)

# pylint: disable=missing-docstring
CONFIG = get_json_config(Settings.path.config.ships)


# pylint: disable=too-many-instance-attributes
class TestShip(Entity):
    # TODO: border? settings?
    def __init__(self, pos, border=None):
        self._type = self.__class__.__name__

        super(TestShip, self).__init__(pos, Settings.path.image.ship.TestShip)

        self._pos = pos
        self._border = border

        self._dx = None
        self._dy = None
        self._fire = False
        self._weapon = None
        self._weapons = InfiniteList()
        self._wbay = None
        self._direction = 0

        self._max_hull = None
        self._max_shield = None
        self._hull = None
        self._shield = None

        # first initialization
        self._load_config(CONFIG[self._type])

    def _load_config(self, config):
        """Load config from mapping."""
        self._dx = int(config.dx)
        self._dy = int(config.dy)
        self._hull = int(config.hull)
        self._shield = int(config.shield)
        self._max_hull = int(config.max_hull)
        self._max_shield = int(config.max_shield)

    @property
    def x(self):
        return self._x

    @property
    def pos(self):
        return self._pos

    def move_left(self):
        self._direction = -1

    def move_right(self):
        self._direction = 1

    def toggle_fire(self):
        self._fire = not self._fire

    def next_weapon(self):
        self._weapons.next()

    def refresh_shield(self):
        pass

    def update(self):
        """Update ship object's state."""
        print("border {0}".format(self._border))
        print("img: <width: {0}><height: {1}>".format(*self._image.get_size()))
        print("pos {0}".format(self._pos))
        print("real {0} : {1}".format(*self.topleft))
        # TODO:
        # think about those who has dx > 1
        if (
                self.x >= self._border.x - self._image.get_width() - 1 and
                self._dx > 0
        ):
            self.x = 0
        elif self.x <= 0 and self._direction < 0:
            self.x = self._border.x - self._image.get_width()

        self.x += self._direction * self._dx
        self._direction = 0

        for weapon in self._weapons:
            weapon.update()

        if self._fire:
            try:
                self._weapon.make_shot(self._pos + self._wbay)
            except ValueError:
                self.next_weapon()

        self.refresh_shield()


# Think about compositing
class Ship(Renderable):
    """Base class for all ships. Contains basic ship logic."""

    compound = True

    def __init__(self, pos, border):
        self._type = self.__class__.__name__
        self._image = None

        self._pos = pos
        self._border = border

        self._dx = None
        self._fire = False
        self._weapon = None
        self._weapons = InfiniteList()
        self._wbay = None
        self._direction = 0

        self._max_hull = None
        self._max_shield = None
        self._hull = None
        self._shield = None

        # Set to True means that object in destroying phase
        # and will ignore remove_obsolete signal
        self._destroy = False

        self._collider = None

        # first initialization
        self._load_config(CONFIG[self._type])

    def _load_config(self, config):
        """Load config from mapping."""

        if not config:
            raise ValueError

        self._dx = int(config.dx)
        self._hull = int(config.hull)
        self._shield = int(config.shield)
        self._max_hull = int(config.max_hull)
        self._max_shield = int(config.max_shield)

    @property
    def pos(self):
        return self._pos

    @property
    def direction(self):
        return self._direction

    @direction.setter
    def direction(self, value):
        if value == 0:
            self._direction = 0
        else:
            self._direction = 1 if value > 0 else -1

    @property
    def max_hull(self):
        return self._max_hull

    @property
    def max_shield(self):
        return self._max_shield

    def get_hull_percentage(self):
        """Return hull percentage."""
        return self._hull * 100.0 / self._max_hull

    def get_shield_percentage(self):
        """Return shield percentage."""
        return self._shield * 100.0 / self._max_shield

    def get_weapon_percentage(self):
        """Return weapon load percentage."""
        return self._weapon.load_percentage()

    def get_render_data(self):
        """Callback for rendering."""
        return [self._pos], self._image.get_image()

    def get_renderable_objects(self):
        """CORP stub."""
        return self._weapons

    def move_left(self):
        """Change direction."""
        self._direction = -1

    def move_right(self):
        """Change direction."""
        self._direction = 1

    def toggle_fire(self):
        """Toggle current weapon fire mode."""
        self._fire = not self._fire

    def next_weapon(self):
        """Select next weapon."""
        self._weapon = self._weapons.next()

    def prev_weapon(self):
        """Select previous weapon."""
        self._weapon = self._weapons.prev()

    def add_weapon(self, weapon):
        """Add new weapon."""

        self._weapons.append(weapon)
        self._weapon = weapon

    def update(self):
        """Update ship object's state."""

        # TODO:
        # think about those who has dx > 1
        if (
                self._pos.x == self._border.x - self._image.width - 1 and
                self._dx > 0
        ):
            self._pos.x = 0
        elif self._pos.x == 1 and self._dx < 0:
            self._pos.x = self._border.x - self._image.width

        self._pos.x += self._direction * self._dx
        self._direction = 0

        for weapon in self._weapons:
            weapon.update()

        if self._fire:
            try:
                # FIXME: [fix-in-place-coercing]
                #        Coercing Point in-place smells badly
                #        (needed now because animation can set float pos to
                #        enemy ships), we need to add type ensurance mechanism
                #        to cast automatically maybe.
                self._weapon.make_shot(self._pos[int] + self._wbay)
            except ValueError:
                self.next_weapon()

        self.refresh_shield()

    def destroy(self):
        """Self-destroying routine."""

        LOG.debug("Destroying ship %s", self)

        if not self._destroy:
            self._destroy = True
            application.get_current().state.collision.remove(self._collider)
            application.get_current().state.remove(self)

    def remove_obsolete(self, pos):
        border = Settings.layout.field.border
        pos = self._pos[int]
        if (
                pos.x > self._image.width + border.x or
                pos.x + self._image.width < 0 or
                int(pos.y) > self._image.height + border.y or
                int(pos.y) + self._image.height < 0
        ) and not self._destroy:
            self.destroy()

    def take_damage(self, damage):
        """Calculate and apply damage to shield and hull."""

        if self._shield < damage:
            rest_damage = damage - self._shield
            self._shield = 0
            self._hull -= rest_damage
        else:
            self._shield -= damage
        if self._hull < 0:
            self._hull = 0

    def refresh_shield(self, amount=1):
        """Refresh shield."""

        if self._shield == self._max_shield:
            return

        if self._shield + amount > self._max_shield:
            self._shield = self._max_shield
        else:
            self._shield += amount


class GenericXEnemy(Ship):
    """Generic X enemy class."""

    def __init__(self, pos, border):
        super(GenericXEnemy, self).__init__(pos, border)
        self._image = Surface([
            "x^x",
            " X ",
            " * ",
        ])

        self._collider = Collider(
            self, [
                "###",
                ".#.",
                ".#.",
            ])

        self.add_weapon(EBlaster())
        self._fire = True
        self._wbay = Point(x=self._image.width // 2, y=1)

        self._animgr = AnimationManager()

    # TODO: rethink this method
    #       Problem of emerging such methods is more complex problem
    #       of creating and configuring GameObjects.
    def add_animation(self, *args, **kwargs):
        self._animgr.add(*args, **kwargs)

    def update(self):
        if self._hull <= 0:
            # TODO: [scoring]
            #       * Parametrize scores, move them to ships.conf
            #       * implement better scoring mechanism
            application.get_current().state.add_player_score(10)
            self.destroy()
            return

        self._animgr.update()

        super(GenericXEnemy, self).update()

    @collision.register("GenericXEnemy", "BasicPlasmaCannon")
    @collision.register("GenericXEnemy", "BasicLaserCharge")
    @collision.register("GenericXEnemy", "BasicUnguidedMissile")
    def collide(self, other, rect):
        self.take_damage(other.damage)
        other.destroy()


class PlayerShip(Ship):
    """PlayerShip class. Contains additional methods for HUD.

    :param :class:`xoinvader.utils.Point` pos: left top corner
    :param :class:`xoinvader.utils.Point` border: max allowed bottom right pos
    """

    def __init__(self, pos, border):
        super(PlayerShip, self).__init__(pos, border)

        self._image = Surface([
            "  O  ",
            "<=H=>",
            " * * ",
        ])

        self._border = border
        self._pos = Point(
            x=pos.x - self._image.width // 2,
            y=pos.y - self._image.height)

        self._collider = Collider(
            self, [
                "  #  ",
                "#####",
                " # # "
            ])

        self._fire = False
        self._weapons = InfiniteList([Blaster(), Laser(), UM()])
        self._weapon = self._weapons.current()
        self._wbay = Point(x=self._image.width // 2, y=-1)

        Mixer().register(
            ".".join([self._type, "engine"]),
            Settings.path.sound.ship[self._type].engine)

    def update(self):
        if self._hull <= 0:
            app = application.get_current()
            app.trigger_state(
                "GameOverState",
                score=app.state.score)

        super(PlayerShip, self).update()

    def get_weapon_info(self):
        """Return information about current weapon."""

        return "Weapon: {w} | [{c}/{m}]".format(
            w=self._weapon.type,
            c=self._weapon.ammo,
            m=self._weapon.max_ammo)

    @collision.register("PlayerShip", "EBasicPlasmaCannon")
    def collide(self, other, rect):
        self.take_damage(other.damage)
        other.destroy()



"""Basic Entity."""
# TODO: better module docstring
# pylint: disable=all

from xoinvader import constants
from xoinvader.common import Settings

if Settings.system.video_driver == constants.DRIVER_SDL:
    import pygame
else:
    class pygame(object):
        Rect = object

        @staticmethod
        def Surface(*args, **kwargs):
            pass


# TODO: document
# TODO: make clear interface
# TODO: component system
class Entity(pygame.Rect):
    """Base class for all game objects."""

    def __init__(self, pos, image=pygame.Surface((0, 0))):
        self._image = pygame.image.load(image)
        super(Entity, self).__init__((pos.x, pos.y), self._image.get_size())

    @property
    def image(self):
        return self._image

    @image.setter
    def image(self, path):
        self._image = pygame.image.load(path)

    def render(self, screen):
        screen.blit(self.image, self.topleft)  # self.move(*map(lambda x: -x, self.image.get_size())))
        pygame.draw.line(screen, (0, 255, 0), (400, 0), (400, 600), 1)
        pygame.draw.line(screen, (0, 255, 0), (0, 300), (800, 300), 1)
        pygame.draw.circle(screen, (0, 0, 255), (400, 300), 100, 1)
        pygame.draw.rect(screen, (255, 0, 0), (0, 0, 800, 600), 1)



"""Module for common shared objects."""

import json
from os.path import dirname

import xoinvader
from xoinvader import constants
from xoinvader.utils import dotdict, Point


def get_json_config(path):
    """Return Settings object made from json.

    :param str path: path to config

    :return xoinvader.utils.dotdict: parsed config
    """

    with open(path) as fd:
        config = dotdict(json.load(fd))
    return config


def rootify(root, config):
    """Append paths in dict to root.

    :param str root: common root
    :param dict config:

    :return dict: rootified config
    """

    for key, path in config.items():
        if isinstance(path, dict):
            rootify(root, path)
        else:
            config[key] = root + path
    return config


__all__ = ["Settings"]

SCREEN_WIDTH = 800
SCREEN_HEIGHT = 600
WIDTH = 90
HEIGHT = 34

_ROOT = dirname(xoinvader.__file__)
_CONFIG = _ROOT + "/config"

DEFAUT_XOI_SETTINGS = dict(
    layout=dict(
        field=dict(
            border=Point(x=WIDTH, y=HEIGHT),
            player=Point(x=WIDTH // 2, y=HEIGHT - 1),
            edge=Point(x=WIDTH, y=HEIGHT - 1)),

        gui=dict(
            bar=dict(
                health=Point(x=2, y=HEIGHT - 1),
                shield=Point(x=22, y=HEIGHT - 1),
                weapon=Point(x=WIDTH - 18, y=HEIGHT - 1)),
            info=dict(
                weapon=Point(x=44, y=HEIGHT - 1)))),
    path=rootify(_ROOT, get_json_config(_ROOT + "/config/path.json")),
    color=dict(
        general=dict(
            normal=None),
        gui=None,
        weapon=None),
    system=dict(
        debug=False,
        encoding=constants.UTF_8,
        no_sound=False,
        no_color=False,
        video_driver=constants.DRIVER_NCURSES,
    )
)

Settings = dotdict(DEFAUT_XOI_SETTINGS)  # pylint: disable=invalid-name


def update_system_settings(args):
    """Update system settings from startup arguments.

    :param dict args: arguments
    """

    for arg, val in args.items():
        if arg in Settings.system:
            Settings.system[arg] = val
        else:
            raise KeyError("No such argument: '{0}'.".format(arg))



"""Module for creating and maintaining xoinvader levels."""


class Level(object):
    """Container for level event sequence and resources.

    Intended to use as just container, but may be subclassed. It's useful in
    order to add methods for creating concrete animations and events, such as
    spawning actual enemies, bonuses, initiating boss fights and so on.

    :param int speed: relative speed of the wave. Means how fast time advances.
    The faster time advances, the shorter the delays between events triggering.
    :param dict events: map from relative time in event to event starting
    function.
    :param bool running: if event sequence currently advances.
    """

    def __init__(self, speed=0):
        self._running = False
        self._counter = 0
        self._speed = speed
        self._events = {}
        self._event_timeouts = []

    @property
    def speed(self):
        """Relative wave's speed.

        :getter: yes
        :setter: yes
        :type: int
        """
        return self._speed

    @speed.setter
    def speed(self, value):
        """Setter."""
        self._speed = value

    @property
    def running(self):
        """If event sequence currently advances.

        :getter: yes
        :setter: no
        :type: bool
        """
        return self._running

    def add_event(self, time, callback):
        """Add event to some point in time.

        :param int time: point in time relative to level start when to run
        `callback`. Callback is fired when `_counter` exceeds provided value
        :param function callback: callback to be fired when wave reaches `time`
        """
        self._events.setdefault(time, []).append(callback)

    def start(self):
        """Start the level.

        Resets the counter, recomputes the list of event timeouts and sets the
        `running` property to `True`.
        """

        self._running = True
        self._counter = 0
        self._event_timeouts = sorted(self._events)

    def update(self):
        """Update the counter and fire appropriate events.

        Function is doing anything useful only when the Level has been started
        earlier, i.e. currently in running state.
        Function updates current timer and fires all events, which timeouts have
        expired. If there's no more events in the queue, the running state
        terminates and object goes to sleep.
        Time counter is increased by `_speed` value at each call of the
        fucntion.
        """

        if not self._running:
            return

        self._counter += self._speed
        while self._event_timeouts[0] <= self._counter:
            current_timeout = self._event_timeouts[0]
            for callback in self._events[current_timeout]:
                callback()
            del self._event_timeouts[0]
            if not self._event_timeouts:
                self._running = False
                break



"""InGameState-specific input/event handlers and commands."""

import curses
import logging
import weakref

from xoinvader import application
from xoinvader.background import Background
from xoinvader.collision import CollisionManager
from xoinvader.common import Settings
from xoinvader.curses_utils import Style
from xoinvader.gui import TextCallbackWidget, TextWidget, WeaponWidget, Bar
from xoinvader.handlers import Handler
from xoinvader.keys import K_A, K_D, K_E, K_F, K_R, K_SPACE, K_ESCAPE, K_Q
from xoinvader.level import Level
from xoinvader.render import render_objects
from xoinvader.ship import GenericXEnemy, PlayerShip
from xoinvader.state import State
from xoinvader.utils import Point, dotdict


LOG = logging.getLogger(__name__)


# pylint: disable=missing-docstring
def move_left_command(actor):
    actor.move_left()


def move_right_command(actor):
    actor.move_right()


def next_weapon_command(actor):
    actor.next_weapon()


def prev_weapon_command(actor):
    actor.prev_weapon()


def toggle_fire_command(actor):
    actor.toggle_fire()


def take_damage_command(actor):
    actor.take_damage(5)


# pylint: disable=protected-access
def switch_actor_command(actor):
    actor._actor, actor._owner.enemy = actor._owner.enemy, actor._actor


def to_mainmenu_command(actor):
    actor.owner.state = "MainMenuState"


# pylint: disable=too-few-public-methods
class InGameInputHandler(Handler):

    def __init__(self, owner):
        super(InGameInputHandler, self).__init__(owner)

        self._command_map = {
            K_A: move_left_command,
            K_D: move_right_command,
            K_E: next_weapon_command,
            K_Q: prev_weapon_command,
            K_R: take_damage_command,
            K_F: switch_actor_command,
            K_SPACE: toggle_fire_command,
            K_ESCAPE: to_mainmenu_command
        }

    def handle(self):
        key = self._screen.getch()
        command = self._command_map.get(key)
        if command:
            if command is to_mainmenu_command:
                command(self._owner)
            elif command is switch_actor_command:
                command(self)
            else:
                command(self._actor)


class InGameEventHandler(Handler):

    def __init__(self, owner):
        super(InGameEventHandler, self).__init__(owner)

        self._input_handler = InGameInputHandler(owner)

    def handle(self):
        self._input_handler.handle()
        # Some other event logic


# pylint: disable=invalid-name
class TestLevel(Level):
    def __init__(self, state_object_adder, speed):
        super(TestLevel, self).__init__(speed)
        self._state_add = state_object_adder

        self._player_ship = PlayerShip(
            Settings.layout.field.player, Settings.layout.field.edge)
        self._state_add(self._player_ship)
        self._enemies = weakref.WeakSet()

        self.add_event(1, self.spawn4)
        self.add_event(100, lambda: None)
        self.add_event(200, self.del4)

        self.bg = Background(Settings.path.level1bg, speed=10, loop=True)
        self.bg.start(True)
        self._state_add(self.bg)

    @staticmethod
    def get_keyframes(y_offset, x_start, direction):
        """Generate keyframe list.

        :param int y_offset: y of top left corner of image
        :param int x_start: x level of left corner of image
        :param int direction: direction of moving: -1 is left, 1 is right
        """

        def new_x(d):
            return x_start - d * direction

        return [
            (0, Point(x_start, y_offset)),
            (4.0, Point(x_start, y_offset + 20)),
            (7.0, Point(new_x(30), y_offset + 10)),
        ]

    def spawn4(self):
        right_side = Settings.layout.field.edge.x

        e1 = GenericXEnemy(
            Point(10, 1),
            Settings.layout.field.edge)
        e2 = GenericXEnemy(
            Point(25, 1),
            Settings.layout.field.edge)

        e3 = GenericXEnemy(
            Point(right_side - 10, 1),
            Settings.layout.field.edge)
        e4 = GenericXEnemy(
            Point(right_side - 25, 1),
            Settings.layout.field.edge)

        e1.add_animation(
            "", e1, "_pos", self.get_keyframes(1, 10, 1), interp=True)
        e2.add_animation(
            "", e2, "_pos", self.get_keyframes(1, 20, 1), interp=True)

        e3.add_animation(
            "", e3, "_pos",
            self.get_keyframes(1, right_side - 10, -1), interp=True)
        e4.add_animation(
            "", e4, "_pos",
            self.get_keyframes(1, right_side - 20, -1), interp=True)

        self._enemies = weakref.WeakSet([e1, e2, e3, e4])
        self._state_add(list(self._enemies))

    def del4(self):
        for enemy in self._enemies:
            enemy.destroy()


class InGameState(State):

    # TODO FIXME: collision manager looks like collision system
    #             belonging to State. Idea is to create some sort of
    #             mechanism for registering/deregistering/accesing
    #             theese systems via GameObject's current State.
    # ATTENTION: Here I met a problem of initialization of this state.
    #            1) here in init we created collision manager (now moved here)
    #            2) initialization of TestLevel created first game object
    #            (PlayerShip) that have collider
    #            3) Collider needs to be registered in CollisionManager that
    #            already was created, BUT! Collider.__init__ can't have
    #            access to current state because several frames above current
    #            control flow resides in State.__init__ and State does not
    #            fully instantiated yet and could not be accessed via
    #            application.state.
    #            One approach that I see here is to create optional postinit
    #            method and create objects that want access to state there.
    # NOTE: I think that here we must extrude systems with such behaviour and
    #       create them before state initialization, maybe here in classfields.
    collision = CollisionManager()

    def __init__(self, owner):
        LOG.info("Instantiating InGame state")
        super(InGameState, self).__init__(owner)
        self._objects = []
        self._screen = self._owner.screen

    def postinit(self):
        """Deferred initialization.

        Prepare GameObjects that require created and registered State object.
        """

        LOG.debug("Registering renderable entities")

        self.level = TestLevel(self.add, speed=1)
        self._actor = self.level._player_ship

        # TODO: [scoring]
        self.score = 0

        self._events = InGameEventHandler(self)

        self.add(self._create_gui())
        self.level.start()

    def add_player_score(self, amount):
        """Add player score.

        :param int amount: scores to add
        """

        self.score += amount

    def get_player_score_string(self):
        """Callback for TextCallbackWidget.

        :return str: score string
        """

        return "Score: {0}".format(self.score)

    # TODO: [object-system]
    #  * implement GameObject common class for using in states
    #  * generalize interaction with game objects and move `add` to base class
    # ATTENTION: renderables that added by another objects in runtime will not
    #  render at the screen, because they must register in state via this func
    #  as others. This is temporary decision as attempt to create playable game
    #  due to deadline.
    def add(self, obj):
        """Add GameObject to State's list of objects.

        Adding via this method means that objects will be
        updated and rendered in main IOLoop.

        :param object obj: GameObject
        """

        obj = obj if isinstance(obj, (list, tuple)) else [obj]
        self._objects.extend(obj)
        LOG.debug("%s", obj)

        # TODO: Because we don't have common GameObject interface
        # This is temporary smellcode
        for item in obj:
            if item.compound:
                subitems = item.get_renderable_objects()
                LOG.debug("Subitems: %s", subitems)
                self._objects.extend(subitems)

    def remove(self, obj):
        """Remove GameObject from State's list of objects.

        Removed objects should be collected by GC.

        :param object obj: GameObject
        """

        LOG.debug("%s", obj)

        try:
            if obj.compound:
                for subobj in obj.get_renderable_objects():
                    self._objects.remove(subobj)
                    del subobj
            self._objects.remove(obj)
        except ValueError:
            LOG.exception("Object %s is not in State's object list.", obj)
        finally:
            del obj

        # TODO: [collider-destruction]
        #       Remove this after collider instant destruction.
        #       Now colliders in weakref set garbage-collected with noticeable
        #       delay. In future we need to drop weakrefs and manually manage
        #       objects.
        LOG.debug("Colliders: %s", len(self.collision._colliders))
        LOG.debug("Collisions: %s", len(self.collision._collisions))
        LOG.debug("Objects in state: %s", len(self._objects))

    def _create_gui(self):
        """Create user interface."""

        comp = dotdict(
            whatever=lambda val: 0.0 <= val <= 100.0,
            high=lambda val: 70.0 <= val <= 100.0,
            normal=lambda val: 35.0 <= val < 70.0,
            low=lambda val: 0.0 <= val < 35.0,
        )

        return [
            Bar(
                pos=Settings.layout.gui.bar.health, prefix="Hull: ",
                general_style=curses.A_BOLD, stylemap={
                    comp.high: Style().gui["dp_ok"],
                    comp.normal: Style().gui["dp_middle"],
                    comp.low: Style().gui["dp_critical"]
                }, callback=self._actor.get_hull_percentage),
            Bar(
                pos=Settings.layout.gui.bar.shield, prefix="Shield: ",
                general_style=curses.A_BOLD, stylemap={
                    comp.high: Style().gui["sh_ok"],
                    comp.normal: Style().gui["sh_mid"],
                    comp.low: Style().gui["dp_critical"]
                }, callback=self._actor.get_shield_percentage),
            Bar(
                pos=Settings.layout.gui.bar.weapon,
                stylemap={
                    comp.whatever: Style().gui["dp_ok"]
                }, callback=self._actor.get_weapon_percentage),
            WeaponWidget(
                Settings.layout.gui.info.weapon,
                self.actor.get_weapon_info)
        ] + [
            TextCallbackWidget(Point(2, 0), self.get_player_score_string),
            TextWidget(
                Point(Settings.layout.field.edge.x // 2 - 4, 0),
                "XOInvader", curses.A_BOLD)
        ]

    def events(self):
        self._events.handle()

    def update(self):
        self.collision.update()
        self.level.update()
        if not self.level.running:
            self.level.start()

        for obj in self._objects:
            obj.update()

    def render(self):
        self._screen.erase()
        self._screen.border(0)

        render_objects(self._objects, self._screen)
        self._screen.refresh()



"""Provides abstract game state."""


class State(object):
    """Represents game state skeleton.

    .. warning:: implement actor base class

    :param owner: state's owner
    :type owner: `xoinvader.application.Application`
    """

    def __init__(self, owner):
        self._owner = owner
        self._actor = None
        self._screen = None

        # TODO:
        self._music = None

    def postinit(self):
        """Do all instantiations that require prepared State object."""

        pass

    def trigger(self, *args, **kwargs):
        """Common way to get useful information for triggered state."""

        pass

    @property
    def owner(self):
        """State's owner.

        :getter: yes
        :setter: no
        :type: :class:`xoinvader.application.Application`
        """
        return self._owner

    @property
    def actor(self):
        """Controllable object.

        :getter: yes
        :setter: no
        :type: object
        """
        return self._actor

    @property
    def screen(self):
        """Screen for rendering state's objects.

        :getter: yes
        :setter: no
        :type: `curses.Window`
        """
        return self._screen

    def events(self):
        "Event handler, calls by `Application.loop` method."
        raise NotImplementedError

    def update(self):
        "Update handler, calls by `Application.loop` method."
        raise NotImplementedError

    def render(self):
        "Render handler, calls by `Application.loop` method."
        raise NotImplementedError



"""XOInvader constants."""


DEFAULT_FPS = 60
"""Default frames per second."""


DRIVER_NCURSES = "ncurses"
"""Ncurses video driver."""

DRIVER_SDL = "pygame-sdl"
"""PyGame SDLv1 video driver."""

# TODO: driver-sdl2; RnD this
DRIVER_SDL2 = "pygame-sdl2"
"""PyGame SDLv2 video driver."""

VIDEO_DRIVERS = [
    DRIVER_NCURSES,
    DRIVER_SDL,
    # DRIVER_SDL2,
]
"""Supported video drivers."""


UTF_8 = "UTF-8"
"""UTF-8 encoding."""



"""
    xoinvader
    ~~~~~~~~~

    XO Invader is an ASCII ncurses-based top-down space shooter.

    :copyright: (c) 2017 by Pavel Kulyov.
    :license: MIT, see LICENSE for more details.
"""

import locale

from .common import Settings
from .common import update_system_settings as _update_system_settings
from .utils import setup_logger


__version__ = "0.1.1"


class XOInitializationError(Exception):
    """Basic initialization exception."""
    pass


def setup_locale(settings):
    """Setup locale settings.

    :param dict settings:

    :return dict: updated settings
    """

    locale.setlocale(locale.LC_ALL, "")
    encoding = locale.getpreferredencoding()
    settings["encoding"] = encoding
    return settings


def init(settings=None):
    """Do engine initialization first.

    :param dict settings:
    """

    if settings is None:
        settings = {}

    try:
        settings = setup_locale(settings)
        log = setup_logger(
            "xoinvader", settings.get("debug", Settings.system.debug))
    except Exception as exc:
        raise XOInitializationError(exc)

    log.debug("Updating engine settings")

    # TODO: config-management: update this stub
    try:
        if settings:
            _update_system_settings(settings)
    except Exception as exc:
        log.exception(exc)
        raise XOInitializationError(exc)


__all__ = ["Settings"]



"""Test xoinvader.animation module."""

import time

import pytest

from xoinvader import animation
from xoinvader.animation import Animation, AnimationManager
from xoinvader.utils import Point, isclose


# pylint: disable=invalid-name,protected-access,missing-docstring
# pylint: disable=too-few-public-methods
class GameObject(object):
    """Game object mock."""

    def __init__(self):
        self.attr = 0
        self.anim = None


@pytest.mark.slow
def test_animation():

    with pytest.raises(ValueError):
        Animation("dummy", bind=None, attr=None, keyframes=[])

    obj = GameObject()
    keyframes = [
        (0.0, 1),
        (0.5, 2),
        (2.0, 10),
    ]
    anim = Animation(name="test", bind=obj, attr="attr", keyframes=keyframes)

    obj.anim = anim
    assert anim.name == "test"

    anim.update()
    assert obj.attr == 1
    time.sleep(1)

    # Frame must be unchanged
    for _ in range(10):
        anim.update()

    assert obj.attr == 2
    time.sleep(1)
    anim.update()  # here time is > 2
    assert obj.attr == 10
    with pytest.raises(StopIteration):
        anim.update()  # It's time to switch animation


@pytest.mark.slow
def test_animation_loop():
    obj = GameObject()
    animgr = AnimationManager()
    animgr.add("test", obj, "attr", keyframes=[(0.0, 0), (1.0, 1)], loop=True)

    animgr.update()
    assert obj.attr == 0
    time.sleep(1)
    animgr.update()
    assert obj.attr == 1
    time.sleep(1)
    animgr.update()
    assert obj.attr == 0
    animgr.update()
    assert obj.attr == 0
    time.sleep(1)
    animgr.update()
    assert obj.attr == 1


@pytest.mark.slow
@pytest.mark.parametrize("loop", (True, False))
def test_animation_interp_loop(loop):
    obj = GameObject()
    animgr = AnimationManager()
    animgr.add(
        "test", obj, "attr", keyframes=[(0.0, 0), (0.4, 1)],
        interp=True, loop=loop)

    animgr.update()
    assert isclose(obj.attr, 0, abs_tol=0.01)
    time.sleep(0.2)
    animgr.update()
    assert isclose(obj.attr, 0.5, abs_tol=0.01)
    time.sleep(0.2)
    animgr.update()
    assert obj.attr == 1
    time.sleep(0.2)
    animgr.update()

    # if not looped - animation stops
    # TODO: test StopIteration, now animgr suppresses it
    if loop:
        assert isclose(obj.attr, 0.5, abs_tol=0.01)
        animgr.update()
    else:
        assert obj.attr == 1
        animgr.update()


@pytest.mark.slow
def test_animation_manager():
    obj = GameObject()
    animgr = AnimationManager()

    with pytest.raises(AttributeError):
        assert animgr.animation

    assert not animgr._animation and animgr.update() is None

    animgr.add("test1", bind=obj, attr="attr", keyframes=[(0.0, 1)])
    animgr.add("test2", bind=obj, attr="attr", keyframes=[(0.0, 2)])

    assert animgr.animation == "test1"

    with pytest.raises(ValueError):
        animgr.animation = "bad-animation"

    animgr.animation = "test2"
    assert animgr.animation == "test2"
    animgr.animation = "test1"

    animgr.update()
    time.sleep(1)

    assert animgr.update() is None


@pytest.mark.parametrize(("values", "types", "expected"), (
    ([1, 2, 3], int, True),
    ([1, 2, "3"], (int, float), False),
    ([1, 2, 3.0], int, False),
    ([1, 2, 3.0], (int, float), True),
))
def test_same_type(values, types, expected):
    assert animation.same_type(values, types) is expected


@pytest.mark.parametrize(("args", "result"), (
    # increasing
    ((0.0, 2.0, 0.0, 1.0, 0.0), 0.0),
    ((0.0, 2.0, 0.0, 1.0, 0.2), 0.4),
    ((0.0, 2.0, 0.0, 1.0, 1.0), 2.0),
    # decreasing
    ((2.0, 1.0, 1.0, 2.0, 1.0), 2.0),
    ((2.0, 1.0, 1.0, 2.0, 1.2), 1.8),
    ((2.0, 1.0, 1.0, 2.0, 2.0), 1.0),
    # negative value
    ((0.0, -1.0, 2.0, 3.0, 2.0), 0.0),
    ((0.0, -1.0, 2.0, 3.0, 2.5), -0.5),
    ((0.0, -1.0, 2.0, 3.0, 3.0), -1.0),
))
def test_linear_equation(args, result):
    assert animation.linear_equation(*args) == result


@pytest.mark.parametrize(("args", "result"), (
    ([(1, 1), (2, 2), 1.5], 1.5),
    ([(1.0, 1.0), (2.0, 2.0), 1.5], 1.5),
    (
        [(1.0, Point(1, 2, 1)), (2.0, Point(2, 1, 2)), 1.5],
        Point(1.5, 1.5, 1.5)
    ),
    (
        [(1.0, Point(1.0, 2.0, 1.0)), (2.0, Point(2.0, 1.0, 2.0)), 1.5],
        Point(1.5, 1.5, 1.5)
    ),
))
def test_interpolate_positive(args, result):
    assert animation.interpolate(*args) == result


@pytest.mark.parametrize(("args", "expected_error"), (
    ([(1.0, 1.0), (2.0, 2.0), 0.99], animation.AnimationBoundariesExceeded),
    ([(1.0, 1.0), (2.0, 2.0), 2.01], animation.AnimationBoundariesExceeded),
    ([(1.0, "a"), (2.0, "b"), 1.5], animation.InterpolationUnknownTypes),
))
def test_interpolate_negative(args, expected_error):
    with pytest.raises(expected_error):
        animation.interpolate(*args)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['xoinvader', 'xoinvader.application', 'xoinvader.tests']

package_data = \
{'': ['*'],
 'xoinvader': ['config/charges.json',
               'config/charges.json',
               'config/charges.json',
               'config/charges.json',
               'config/path.json',
               'config/path.json',
               'config/path.json',
               'config/path.json',
               'config/ships.json',
               'config/ships.json',
               'config/ships.json',
               'config/ships.json',
               'config/weapons.json',
               'config/weapons.json',
               'config/weapons.json',
               'config/weapons.json',
               'res/*',
               'res/gfx/*',
               'res/snd/*'],
 'xoinvader.tests': ['fixtures/chunk_duplicate_names.bg',
                     'fixtures/chunk_duplicate_names.bg',
                     'fixtures/chunk_duplicate_names.bg',
                     'fixtures/chunk_duplicate_names.bg',
                     'fixtures/chunk_duplicate_names.bg',
                     'fixtures/chunk_duplicate_names.bg',
                     'fixtures/chunk_duplicate_names.bg',
                     'fixtures/chunk_no_name.bg',
                     'fixtures/chunk_no_name.bg',
                     'fixtures/chunk_no_name.bg',
                     'fixtures/chunk_no_name.bg',
                     'fixtures/chunk_no_name.bg',
                     'fixtures/chunk_no_name.bg',
                     'fixtures/chunk_no_name.bg',
                     'fixtures/chunk_normal.bg',
                     'fixtures/chunk_normal.bg',
                     'fixtures/chunk_normal.bg',
                     'fixtures/chunk_normal.bg',
                     'fixtures/chunk_normal.bg',
                     'fixtures/chunk_normal.bg',
                     'fixtures/chunk_normal.bg',
                     'fixtures/no_chunks.bg',
                     'fixtures/no_chunks.bg',
                     'fixtures/no_chunks.bg',
                     'fixtures/no_chunks.bg',
                     'fixtures/no_chunks.bg',
                     'fixtures/no_chunks.bg',
                     'fixtures/no_chunks.bg',
                     'fixtures/scoreboard_corrupted',
                     'fixtures/scoreboard_corrupted',
                     'fixtures/scoreboard_corrupted',
                     'fixtures/scoreboard_corrupted',
                     'fixtures/scoreboard_corrupted',
                     'fixtures/scoreboard_corrupted',
                     'fixtures/scoreboard_corrupted',
                     'fixtures/scoreboard_empty',
                     'fixtures/scoreboard_empty',
                     'fixtures/scoreboard_empty',
                     'fixtures/scoreboard_empty',
                     'fixtures/scoreboard_empty',
                     'fixtures/scoreboard_empty',
                     'fixtures/scoreboard_empty',
                     'fixtures/scoreboard_valid',
                     'fixtures/scoreboard_valid',
                     'fixtures/scoreboard_valid',
                     'fixtures/scoreboard_valid',
                     'fixtures/scoreboard_valid',
                     'fixtures/scoreboard_valid',
                     'fixtures/scoreboard_valid']}

install_requires = \
['pygame', 'tornado>=4,<5']

setup_kwargs = {
    'name': 'xoinvader',
    'version': '0.1.1',
    'description': '(not so) small python curses space game',
    'long_description': None,
    'author': 'Pavel Kulyov',
    'author_email': 'kulyov.pavel@gmail.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3,<4',
}


setup(**setup_kwargs)



"""Test xoinvader.__init__ module."""

import pytest

import xoinvader
from xoinvader.common import Settings
from xoinvader.utils import dotdict


# pylint: disable=missing-docstring
def test_init():
    Settings.system.test_section = dotdict(key="value")
    xoinvader.init()
    xoinvader.init({"test_section": dotdict(another_key="another-value")})
    assert Settings.system.test_section.another_key == "another-value"

    with pytest.raises(xoinvader.XOInitializationError):
        xoinvader.init("bad-value")

    with pytest.raises(xoinvader.XOInitializationError):
        xoinvader.init({"bad-key": "value"})



"""Test xoinvader.utils module."""

import pytest

from xoinvader.utils import (
    setup_logger,
    dotdict,
    isclose,
    InfiniteList,
    Point,
    Surface,
    Timer,
)


# pylint: disable=invalid-name,protected-access,missing-docstring
def test_setup_logger():
    logger = setup_logger("test", True)
    assert logger


def test_dotdict_setattr():
    settings = dotdict()
    settings.test_entry = 42
    assert settings["test_entry"] == 42
    assert settings.test_entry == 42

    settings["test_entry_2"] = 42
    assert settings.test_entry == 42

    assert pytest.raises(AttributeError, lambda: settings.bad_key)


@pytest.mark.parametrize(("left", "right", "kwargs", "expected"), (
    (0.0, 0.0, {}, True),
    (1.0, 0.0, {}, False),
    (0.1, 0.0, {}, False),
    (0.1, 0.0, {"abs_tol": 0.1}, True),
    (0.1, 0.0, {"rel_tol": 0.1}, False),
))
def test_isclose(left, right, kwargs, expected):
    """xoinvader.utils.isclose"""
    assert isclose(left, right, **kwargs) is expected


def test_point_operations():
    ax, ay, az = 10, 10, 10
    bx, by, bz = 20, 20, 20
    a = Point(ax, ay, az)
    b = Point(bx, by, bz)

    assert a.x == ax
    assert a.y == ay
    assert a.z == az
    assert b.x == bx
    assert b.y == by
    assert b.z == bz

    assert repr(a) == "Point(x={0}, y={1}, z={2})".format(
        a.x, a.y, a.z)

    assert a + b == Point(ax + bx, ay + by, az + bz)
    assert a - b == Point(ax - bx, ay - by, az - bz)
    assert a + 5 == Point(ax + 5, ay + 5, az + 5)
    assert a - 5 == Point(ax - 5, ay - 5, az - 5)
    assert a * 5 == Point(ax * 5, ay * 5, az * 5)
    assert a / 5 == Point(ax / 5, ay / 5, az / 5)

    with pytest.raises(ValueError):
        assert a + "a"
    with pytest.raises(ValueError):
        assert a - "a"
    with pytest.raises(ValueError):
        assert a * "a"
    with pytest.raises(ValueError):
        assert a / "a"
    with pytest.raises(ValueError):
        assert a == "a"

    a.x = bx
    a.y = by
    a.z = bz

    assert a.x == bx
    assert a.y == by
    assert a.z == bz

    b.x = -bx
    b.y = -by
    b.z = -bz

    assert a + b == Point(0, 0, 0)
    assert a + Point(-50, -50, -50) == Point(-30, -30, -30)

    assert Point(1.9, 1.9)[int] == Point(1, 1, 0)


def test_infinite_list_operations():
    # Test empty InfiniteList behaviour

    inf_list = InfiniteList()
    for func in [inf_list.current, inf_list.next, inf_list.prev]:
        with pytest.raises(IndexError):
            func()

    # Test one element behaviour
    data = "test1"
    inf_list = InfiniteList([data])

    assert len(inf_list) == 1
    assert inf_list[0] == data
    assert inf_list.current() == data
    assert inf_list.next() == data
    assert inf_list.prev() == data

    # Test many elements behaviour


_image = [
    [" ", "O", " "],
    ["x", "X", "x"]]


def test_surface_attributes():
    surface = Surface(_image)
    assert surface.height == len(_image)
    assert surface.width == len(_image[0])
    assert surface.raw == _image


def test_image_generator():
    surface = Surface(_image)
    image_gen = surface.get_image()
    for lpos, image, style in image_gen:
        assert _image[lpos.y][lpos.x] == image
        assert style is None


@pytest.mark.slow
def test_timer_get_elapsed():
    timer = Timer(5.0, lambda: True)
    timer.start()
    while timer.running:
        assert timer.get_elapsed() >= 0.0
        timer.update()
        assert timer.get_elapsed() >= 0.0



"""Tests for xoinvader.state module."""


import pytest

from xoinvader.state import State


def test_state():
    """Test xoinvader.state.State base class."""

    state = State("owner")

    assert state.owner == "owner"
    assert state.actor is None
    assert state.screen is None

    with pytest.raises(NotImplementedError):
        state.events()
    with pytest.raises(NotImplementedError):
        state.update()
    with pytest.raises(NotImplementedError):
        state.render()



"""Rendering routines."""


from abc import ABCMeta, abstractmethod
from operator import attrgetter

from six import add_metaclass

from xoinvader.common import Settings
from xoinvader.utils import Point


INVISIBLE_SYMBOLS = " "
"""Symbols that renderer must not render on the screen."""


@add_metaclass(ABCMeta)
class Renderable(object):
    """Base for renderable objects.

    .. class-variables::

    * compound (bool): if object consists of other renderables

    * render_priority (int): priority for renderer, greater -> rendered later

    * draw_on_border (bool): allow or not drawing on border
    if not allowed - renderer will send remove_obsolete signal to object
    """

    compound = False
    render_priority = 0
    draw_on_border = False

    @abstractmethod
    def get_render_data(self):
        """Renderable.get_render_data(None) -> (gpos_list, data_gen)

        Every renderable object must return tuple consist of:
            * gpos_list: list of every Surface's global positions [Points]
            Example: [Point(x=5, y=5), Point(x=10, y=10)]

            * data_gen: generator which yields tuple (lpos, image, style)
            Example: (Point(x=5, y=5), "*", curses.A_BOLD)
        """
        pass

    @classmethod
    def type(cls):
        return cls.__name__

    def remove_obsolete(self, pos):
        """Renderable.remove_obsolete(Point(int, int)) -> None

        Every renderable object must remove old bullets that places behind
        border (field for rendering).

        If object will never change its coordinates it may not implement this
        method.
        """
        pass


# TODO: [advanced-render]
# * use render for curses and SDL windows
def render_objects(objects, screen):
    """Render all renderable objects.

    :param collections.Iterable objects: objects to render
    :param curses._Window screen: curses screen
    """
    border = Point(*screen.getmaxyx()[::-1])

    # TODO: Move sorting to some kind of object manager
    # Must be sorted on adding objects
    for obj in sorted(objects, key=attrgetter("render_priority")):
        gpos_list, data_gen = obj.get_render_data()

        for data in data_gen:
            for gpos in gpos_list:
                lpos, image, style = data
                cpos = (gpos + lpos)[int]

                if (
                        (cpos.x >= border.x - 1 or cpos.y >= border.y - 1) or
                        (cpos.x <= 0 or cpos.y <= 0)
                ) and not obj.draw_on_border:
                    obj.remove_obsolete(gpos)
                    continue

                if image in INVISIBLE_SYMBOLS:
                    continue

                if style:
                    screen.addstr(
                        cpos.y, cpos.x,
                        image.encode(Settings.system.encoding), style)
                else:
                    screen.addstr(cpos.y, cpos.x, image)
